"""Base Synthesizer class."""

import datetime
import functools
import inspect
import logging
import math
import os
import uuid
import warnings
from collections import defaultdict

import cloudpickle
import copulas
import numpy as np
import pandas as pd
import pkg_resources
import tqdm
from copulas.multivariate import GaussianMultivariate

from sdv.data_processing.data_processor import DataProcessor
from sdv.errors import ConstraintsNotMetError, SynthesizerInputError
from sdv.single_table.errors import InvalidDataError
from sdv.single_table.utils import check_num_rows, handle_sampling_error, validate_file_path
from sdv.utils import is_boolean_type, is_datetime_type, is_numerical_type

LOGGER = logging.getLogger(__name__)
COND_IDX = str(uuid.uuid4())
FIXED_RNG_SEED = 73251
TMP_FILE_NAME = '.sample.csv.temp'
DISABLE_TMP_FILE = 'disable'


class BaseSynthesizer:
    """Base class for all ``Synthesizers``.

    The ``BaseSynthesizer`` class defines the common API that all the
    ``Synthesizers`` need to implement, as well as common functionality.

    Args:
        metadata (sdv.metadata.SingleTableMetadata):
            Single table metadata representing the data that this synthesizer will be used for.
        enforce_min_max_values (bool):
            Specify whether or not to clip the data returned by ``reverse_transform`` of
            the numerical transformer, ``FloatFormatter``, to the min and max values seen
            during ``fit``. Defaults to ``True``.
        enforce_rounding (bool):
            Define rounding scheme for ``numerical`` columns. If ``True``, the data returned
            by ``reverse_transform`` will be rounded as in the original data. Defaults to ``True``.
    """

    _model_sdtype_transformers = None

    def _validate_inputs(self, enforce_min_max_values, enforce_rounding):
        if not isinstance(enforce_min_max_values, bool):
            raise SynthesizerInputError(
                f"Invalid value '{enforce_min_max_values}' for parameter 'enforce_min_max_values'."
                ' Please provide True or False.'
            )

        if not isinstance(enforce_rounding, bool):
            raise SynthesizerInputError(
                f"Invalid value '{enforce_rounding}' for parameter 'enforce_rounding'."
                ' Please provide True or False.'
            )

    def _update_default_transformers(self):
        if self._model_sdtype_transformers is not None:
            for sdtype, transformer in self._model_sdtype_transformers.items():
                self._data_processor._update_transformers_by_sdtypes(sdtype, transformer)

    def __init__(self, metadata, enforce_min_max_values=True, enforce_rounding=True):
        self._validate_inputs(enforce_min_max_values, enforce_rounding)
        self.metadata = metadata
        self.metadata.validate()
        self.enforce_min_max_values = enforce_min_max_values
        self.enforce_rounding = enforce_rounding
        self._data_processor = DataProcessor(
            metadata=self.metadata,
            enforce_rounding=self.enforce_rounding,
            enforce_min_max_values=self.enforce_min_max_values
        )
        self._fitted = False
        self._random_state_set = False
        self._update_default_transformers()
        self._creation_date = datetime.datetime.today().strftime('%Y-%m-%d')
        self._fitted_date = None
        self._fitted_sdv_version = None

    def _validate_metadata_matches_data(self, columns):
        errors = []
        metadata_columns = self.metadata.columns or []
        missing_data_columns = set(columns).difference(metadata_columns)
        if missing_data_columns:
            errors.append(
                f'The columns {sorted(missing_data_columns)} are not present in the metadata.')

        missing_metadata_columns = set(metadata_columns).difference(columns)
        if missing_metadata_columns:
            errors.append(
                f'The metadata columns {sorted(missing_metadata_columns)} '
                'are not present in the data.'
            )

        if errors:
            raise InvalidDataError(errors)

    def _get_primary_and_alternate_keys(self):
        keys = set(self.metadata.alternate_keys)
        if self.metadata.primary_key:
            keys.update({self.metadata.primary_key})

        return keys

    def _get_set_of_sequence_keys(self):
        if isinstance(self.metadata.sequence_key, tuple):
            return set(self.metadata.sequence_key)

        if isinstance(self.metadata.sequence_key, str):
            return {self.metadata.sequence_key}

        return set()

    def _validate_keys_dont_have_missing_values(self, data):
        errors = []
        keys = self._get_primary_and_alternate_keys()
        keys.update(self._get_set_of_sequence_keys())
        for key in sorted(keys):
            if pd.isna(data[key]).any():
                errors.append(f"Key column '{key}' contains missing values.")

        return errors

    @staticmethod
    def _format_invalid_values_string(invalid_values):
        invalid_values = sorted(invalid_values, key=lambda x: str(x))
        if len(invalid_values) > 3:
            return invalid_values[:3] + [f'+ {len(invalid_values) - 3} more']

        return invalid_values

    def _validate_key_values_are_unique(self, data):
        errors = []
        keys = self._get_primary_and_alternate_keys()
        for key in sorted(keys):
            repeated_values = set(data[key][data[key].duplicated()])
            if repeated_values:
                repeated_values = self._format_invalid_values_string(repeated_values)
                errors.append(f"Key column '{key}' contains repeating values: {repeated_values}")

        return errors

    def _validate_sdtype(self, sdtype, column, validation):
        valid = column.apply(validation)
        invalid_values = set(column[~valid])
        if invalid_values:
            invalid_values = self._format_invalid_values_string(invalid_values)
            return [f"Invalid values found for {sdtype} column '{column.name}': {invalid_values}."]

        return []

    def _validate_column(self, column):
        """Validate values of the column satisfy its sdtype properties."""
        errors = []
        sdtype = self.metadata.columns[column.name]['sdtype']

        # boolean values must be True/False, None or missing values
        # int/str are not allowed
        if sdtype == 'boolean':
            errors += self._validate_sdtype(sdtype, column, is_boolean_type)

        # numerical values must be int/float, None or missing values
        # str/bool are not allowed
        if sdtype == 'numerical':
            errors += self._validate_sdtype(sdtype, column, is_numerical_type)

        # datetime values must be castable to datetime, None or missing values
        if sdtype == 'datetime':
            errors += self._validate_sdtype(
                sdtype, column, lambda x: pd.isna(x) | is_datetime_type(x))

        return errors

    def _validate(self, data):
        """Validate any rules that only apply to specific synthesizers.

        This method should be overridden by subclasses.
        """
        return []

    def validate(self, data):
        """Validate data.

        Args:
            data (pd.DataFrame):
                The data to validate.

        Raises:
            ValueError:
                Raised when data is not of type pd.DataFrame.
            InvalidDataError:
                Raised if:
                    * data columns don't match metadata
                    * keys have missing values
                    * primary or alternate keys are not unique
                    * context columns vary for a sequence key
                    * values of a column don't satisfy their sdtype
        """
        if not isinstance(data, pd.DataFrame):
            raise ValueError(f'Data must be a DataFrame, not a {type(data)}.')

        # Both metadata and data must have the same set of columns
        self._validate_metadata_matches_data(data.columns)

        errors = []
        # Primary, sequence and alternate keys can't have missing values
        errors += self._validate_keys_dont_have_missing_values(data)

        # Primary and alternate key values must be unique
        errors += self._validate_key_values_are_unique(data)

        # Any other rules that must be met
        errors += self._validate(data)

        # Every column must satisfy the properties of their sdtypes
        for column in data:
            errors += self._validate_column(data[column])

        if errors:
            raise InvalidDataError(errors)

    def _validate_transformers(self, column_name_to_transformer):
        keys = self._get_primary_and_alternate_keys() | self._get_set_of_sequence_keys()
        for column, transformer in column_name_to_transformer.items():
            if transformer is None:
                continue

            if column in keys and not transformer.is_generator():
                raise SynthesizerInputError(
                    f"Column '{column}' is a key. It cannot be preprocessed using "
                    f"the '{type(transformer).__name__}' transformer."
                )

            # If columns were set, the transformer was fitted
            if transformer.columns:
                raise SynthesizerInputError(
                    f"Transformer for column '{column}' has already been fit on data.")

    def _warn_for_update_transformers(self, column_name_to_transformer):
        """Raise warnings for update_transformers.

        Args:
            column_name_to_transformer (dict):
                Dict mapping column names to transformers to be used for that column.
        """
        for column in column_name_to_transformer:
            sdtype = self.metadata.columns[column]['sdtype']
            if sdtype in {'categorical', 'boolean'}:
                warnings.warn(
                    f"Replacing the default transformer for column '{column}' "
                    'might impact the quality of your synthetic data.'
                )

    def update_transformers(self, column_name_to_transformer):
        """Update any of the transformers assigned to each of the column names.

        Args:
            column_name_to_transformer (dict):
                Dict mapping column names to transformers to be used for that column.
        """
        self._validate_transformers(column_name_to_transformer)
        self._warn_for_update_transformers(column_name_to_transformer)
        self._data_processor.update_transformers(column_name_to_transformer)
        if self._fitted:
            msg = 'For this change to take effect, please refit the synthesizer using `fit`.'
            warnings.warn(msg, UserWarning)

    def get_parameters(self):
        """Return the parameters used to instantiate the synthesizer."""
        parameters = inspect.signature(self.__init__).parameters
        instantiated_parameters = {}
        for parameter_name in parameters:
            if parameter_name != 'metadata':
                instantiated_parameters[parameter_name] = self.__dict__.get(parameter_name)

        return instantiated_parameters

    def get_metadata(self):
        """Return the ``SingleTableMetadata`` for this synthesizer."""
        return self.metadata

    def load_custom_constraint_classes(self, filepath, class_names):
        """Load a custom constraint class for the current synthesizer.

        Args:
            filepath (str):
                String representing the absolute or relative path to the python file where
                the custom constraints are declared.
            class_names (list):
                A list of custom constraint classes to be imported.
        """
        self._data_processor.load_custom_constraint_classes(filepath, class_names)

    def add_custom_constraint_class(self, class_object, class_name):
        """Add a custom constraint class for the synthesizer to use.

        Args:
            class_object (sdv.constraints.Constraint):
                A custom constraint class object.
            class_name (str):
                The name to assign this custom constraint class. This will be the name to use
                when writing a constraint dictionary for ``add_constraints``.
        """
        self._data_processor.add_custom_constraint_class(class_object, class_name)

    def add_constraints(self, constraints):
        """Add constraints to the synthesizer.

        Args:
            constraints (list):
                List of constraints described as dictionaries in the following format:
                    * ``constraint_class``: Name of the constraint to apply.
                    * ``constraint_parameters``: A dictionary with the constraint parameters.
        """
        if self._fitted:
            warnings.warn(
                "For these constraints to take effect, please refit the synthesizer using 'fit'."
            )

        self._data_processor.add_constraints(constraints)

    def get_constraints(self):
        """Get a list of the current constraints that will be used.

        Returns:
            list:
                List of dictionaries describing the constraints for this synthesizer.
        """
        return self._data_processor.get_constraints()

    def auto_assign_transformers(self, data):
        """Automatically assign the required transformers for the given data and constraints.

        This method will automatically set a configuration to the ``rdt.HyperTransformer``
        with the required transformers for the current data.

        Args:
            data (pandas.DataFrame):
                The raw data (before any transformations) that will be used to fit the model.
        """
        self._data_processor.prepare_for_fitting(data)

    def get_transformers(self):
        """Get a dictionary mapping of ``column_name``  and ``rdt.transformers``.

        A dictionary representing the column names and the transformers that will be used
        to transform the data.

        Returns:
            dict:
                A dictionary mapping with column names and transformers.
        """
        field_transformers = self._data_processor._hyper_transformer.field_transformers
        if field_transformers == {}:
            raise ValueError(
                "No transformers were returned in 'get_transformers'. "
                "Use 'auto_assign_transformers' or 'fit' to create them."
            )

        # Order the output to match metadata
        ordered_field_transformers = {
            column_name: field_transformers.get(column_name)
            for column_name in self.metadata.columns
            if column_name in field_transformers
        }

        # Add missing columns created by the constraints
        ordered_field_transformers.update(field_transformers)

        return ordered_field_transformers

    def get_info(self):
        """Get dictionary with information regarding the synthesizer.

        Return:
            dict:
                * ``class_name``: synthesizer class name
                * ``creation_date``: date of creation
                * ``is_fit``: whether or not the synthesizer has been fit
                * ``last_fit_date``: date for the last time it was fit
                * ``fitted_sdv_version``: version of sdv it was on when fitted
        """
        return {
            'class_name': self.__class__.__name__,
            'creation_date': self._creation_date,
            'is_fit': self._fitted,
            'last_fit_date': self._fitted_date,
            'fitted_sdv_version': self._fitted_sdv_version
        }

    def _preprocess(self, data):
        self.validate(data)
        self._data_processor.fit(data)
        return self._data_processor.transform(data)

    def preprocess(self, data):
        """Transform the raw data to numerical space.

        Args:
            data (pandas.DataFrame):
                The raw data to be transformed.

        Returns:
            pandas.DataFrame:
                The preprocessed data.
        """
        if self._fitted:
            warnings.warn(
                'This model has already been fitted. To use the new preprocessed data, '
                "please refit the model using 'fit' or 'fit_processed_data'."
            )

        return self._preprocess(data)

    def _fit(self, processed_data):
        """Fit the model to the table.

        Args:
            processed_data (pandas.DataFrame):
                Data to be learned.
        """
        raise NotImplementedError()

    def fit_processed_data(self, processed_data):
        """Fit this model to the transformed data.

        Args:
            processed_data (pandas.DataFrame):
                The transformed data used to fit the model to.
        """
        self._fit(processed_data)
        self._fitted = True
        self._fitted_date = datetime.datetime.today().strftime('%Y-%m-%d')
        self._fitted_sdv_version = pkg_resources.get_distribution('sdv').version

    def fit(self, data):
        """Fit this model to the original data.

        Args:
            data (pandas.DataFrame):
                The raw data (before any transformations) to fit the model to.
        """
        self._fitted = False
        self._data_processor.reset_sampling()
        self._random_state_set = False
        processed_data = self._preprocess(data)
        self.fit_processed_data(processed_data)

    def save(self, filepath):
        """Save this model instance to the given path using cloudpickle.

        Args:
            filepath (str):
                Path where the synthesizer instance will be serialized.
        """
        with open(filepath, 'wb') as output:
            cloudpickle.dump(self, output)

    @classmethod
    def load(cls, filepath):
        """Load a TabularModel instance from a given path.

        Args:
            filepath (str):
                Path from which to load the serialized synthesizer.

        Returns:
            SingleTableSynthesizer:
                The loaded synthesizer.
        """
        with open(filepath, 'rb') as f:
            model = cloudpickle.load(f)
            return model


class BaseSingleTableSynthesizer(BaseSynthesizer):
    """Base class for all single-table ``Synthesizers``.

    The ``BaseSingleTableSynthesizer`` class defines the common sampling methods
    for all single-table synthesizers.
    """

    def _set_random_state(self, random_state):
        """Set the random state of the model's random number generator.

        Args:
            random_state (int, tuple[np.random.RandomState, torch.Generator], or None):
                Seed or tuple of random states to use.
        """
        self._model.set_random_state(random_state)
        self._random_state_set = True

    def reset_sampling(self):
        """Reset the sampling to the state that was left right after fitting."""
        self._data_processor.reset_sampling()
        self._random_state_set = False

    @staticmethod
    def _filter_conditions(sampled, conditions, float_rtol):
        """Filter the sampled rows that match the conditions.

        If condition columns are float values, consider a match anything that
        is closer than the given ``float_rtol`` and then make the value exact.

        Args:
            sampled (pandas.DataFrame):
                The sampled rows, reverse transformed.
            conditions (dict):
                The dictionary of conditioning values.
            float_rtol (float):
                Maximum tolerance when considering a float match.

        Returns:
            pandas.DataFrame:
                Rows from the sampled data that match the conditions.
        """
        for column, value in conditions.items():
            column_values = sampled[column]
            if column_values.dtype.kind == 'f':
                distance = value * float_rtol
                sampled = sampled[np.abs(column_values - value) <= distance]
                sampled[column] = value
            else:
                sampled = sampled[column_values == value]

        return sampled

    def _sample_rows(self, num_rows, conditions=None, transformed_conditions=None,
                     float_rtol=0.1, previous_rows=None):
        """Sample rows with the given conditions.

        Input conditions is taken both in the raw input format, which will be used
        for filtering during the reject-sampling loop, and already transformed
        to the model format, which will be passed down to the model if it supports
        conditional sampling natively.

        If condition columns are float values, consider a match anything that
        is closer than the given ``float_rtol`` and then make the value exact.

        If the model does not have any data columns, the result of this call
        is a dataframe of the requested length with no columns in it.

        If there are no columns other than the ``primary_key``, this will proceed to sample
        only the ``primary_key`` using the ``DataProcessor``.

        Args:
            num_rows (int):
                Number of rows to sample.
            conditions (dict):
                The dictionary of conditioning values in the original format.
            transformed_conditions (dict):
                The dictionary of conditioning values transformed to the model format.
            float_rtol (float):
                Maximum tolerance when considering a float match.
            previous_rows (pandas.DataFrame):
                Valid rows sampled in the previous iterations.

        Returns:
            tuple:
                * pandas.DataFrame:
                    Rows from the sampled data that match the conditions.
                * int:
                    Number of rows that are considered valid.
        """
        if not self._random_state_set:
            self._set_random_state(FIXED_RNG_SEED)

        if self._data_processor.get_sdtypes(primary_keys=False):
            if conditions is None:
                sampled = self._sample(num_rows)
            else:
                try:
                    sampled = self._sample(num_rows, transformed_conditions)
                except NotImplementedError:
                    sampled = self._sample(num_rows)

            sampled = self._data_processor.reverse_transform(sampled)

            if previous_rows is not None:
                sampled = pd.concat([previous_rows, sampled], ignore_index=True)

            sampled = self._data_processor.filter_valid(sampled)

            if conditions is not None:
                sampled = self._filter_conditions(sampled, conditions, float_rtol)

            num_valid = len(sampled)

            return sampled, num_valid

        else:
            sampled = pd.DataFrame(index=range(num_rows))
            sampled = self._data_processor.reverse_transform(sampled)
            return sampled, num_rows

    def _sample_batch(self, batch_size, max_tries=100,
                      conditions=None, transformed_conditions=None, float_rtol=0.01,
                      progress_bar=None, output_file_path=None):
        """Sample a batch of rows with the given conditions.

        This will enter a reject-sampling loop in which rows will be sampled until
        all of them are valid and match the requested conditions. If ``max_tries``
        is exceeded, it will return as many rows as it has sampled, which may be less
        than the target number of rows.

        Input conditions is taken both in the raw input format, which will be used
        for filtering during the reject-sampling loop, and already transformed
        to the model format, which will be passed down to the model if it supports
        conditional sampling natively.

        If condition columns are float values, consider a match anything that is
        relatively closer than the given ``float_rtol`` and then make the value exact.

        If the model does not have any data columns, the result of this call
        is a dataframe of the requested length with no columns in it.

        Args:
            batch_size (int):
                Number of rows to sample for this batch. If not given the model
                will generate as many rows as there were in the
                data passed to the ``fit`` method.
            max_tries (int):
                Number of times to retry sampling until the batch size is met.
                Defaults to 100.
            conditions (dict):
                The dictionary of conditioning values in the original input format.
            transformed_conditions (dict):
                The dictionary of conditioning values transformed to the model format.
            float_rtol (float):
                Maximum tolerance when considering a float match.
            progress_bar (tqdm.tqdm or None):
                The progress bar to update when sampling. If None, a new tqdm progress
                bar will be created.
            output_file_path (str or None):
                The file to periodically write sampled rows to. If None, does not write
                rows anywhere.

        Returns:
            pandas.DataFrame:
                Sampled data.
        """
        num_rows_to_sample = batch_size

        counter = 0
        num_valid = 0
        prev_num_valid = None
        remaining = batch_size
        sampled = pd.DataFrame()

        while num_valid < batch_size and counter < max_tries:
            prev_num_valid = num_valid
            sampled, num_valid = self._sample_rows(
                num_rows_to_sample,
                conditions,
                transformed_conditions,
                float_rtol,
                sampled
            )

            num_new_valid_rows = num_valid - prev_num_valid
            num_increase = min(num_new_valid_rows, remaining)
            num_sampled = min(len(sampled), batch_size)
            if num_increase > 0:
                if output_file_path:
                    append_kwargs = {'mode': 'a', 'header': False}
                    append_kwargs = append_kwargs if os.path.getsize(output_file_path) > 0 else {}
                    sampled.head(num_sampled).tail(num_increase).to_csv(
                        output_file_path,
                        index=False,
                        **append_kwargs,
                    )

                if progress_bar is not None:
                    progress_bar.update(num_increase)

            remaining = batch_size - num_valid
            valid_rate = max(num_new_valid_rows, 1) / max(num_rows_to_sample, 1)
            num_rows_to_sample = min(10 * batch_size, int(remaining / valid_rate))

            if remaining > 0:
                LOGGER.info(
                    f'{remaining} valid rows remaining. Resampling {num_rows_to_sample} rows')

            counter += 1

        return sampled.head(min(len(sampled), batch_size))

    @staticmethod
    def _make_condition_dfs(conditions):
        """Transform ``conditions`` into a list of dataframes.

        Args:
            conditions (list[sdv.sampling.Condition]):
                A list of ``sdv.sampling.Condition``, where each ``Condition`` object
                represents a desired column value mapping and the number of rows
                to generate for that condition.

        Returns:
            list[pandas.DataFrame]:
                A list of ``conditions`` as dataframes.
        """
        condition_dataframes = defaultdict(list)
        for condition in conditions:
            column_values = condition.get_column_values()
            condition_dataframes[tuple(column_values.keys())].append(
                pd.DataFrame(column_values, index=range(condition.get_num_rows())))

        return [
            pd.concat(condition_list, ignore_index=True)
            for condition_list in condition_dataframes.values()
        ]

    def _sample_in_batches(self, num_rows, batch_size, max_tries_per_batch, conditions=None,
                           transformed_conditions=None, float_rtol=0.01, progress_bar=None,
                           output_file_path=None):
        sampled = []
        batch_size = batch_size if num_rows > batch_size else num_rows
        for step in range(math.ceil(num_rows / batch_size)):
            sampled_rows = self._sample_batch(
                batch_size=batch_size,
                max_tries=max_tries_per_batch,
                conditions=conditions,
                transformed_conditions=transformed_conditions,
                float_rtol=float_rtol,
                progress_bar=progress_bar,
                output_file_path=output_file_path,
            )
            sampled.append(sampled_rows)

        sampled = pd.concat(sampled, ignore_index=True) if len(sampled) > 0 else pd.DataFrame()
        return sampled.head(num_rows)

    def _conditionally_sample_rows(self, dataframe, condition, transformed_condition,
                                   max_tries_per_batch=None, batch_size=None, float_rtol=0.01,
                                   graceful_reject_sampling=True, progress_bar=None,
                                   output_file_path=None):
        batch_size = batch_size or len(dataframe)
        sampled_rows = self._sample_in_batches(
            num_rows=len(dataframe),
            batch_size=batch_size,
            max_tries_per_batch=max_tries_per_batch,
            conditions=condition,
            transformed_conditions=transformed_condition,
            float_rtol=float_rtol,
            progress_bar=progress_bar,
            output_file_path=output_file_path
        )

        if len(sampled_rows) > 0:
            sampled_rows[COND_IDX] = dataframe[COND_IDX].to_numpy()[:len(sampled_rows)]

        elif not graceful_reject_sampling:
            user_msg = (
                'Unable to sample any rows for the given conditions '
                f"'{transformed_condition}'. "
            )
            if hasattr(self, '_model') and isinstance(self._model, GaussianMultivariate):
                user_msg = user_msg + (
                    'This may be because the provided values are out-of-bounds in the '
                    'current model. \nPlease try again with a different set of values.'
                )
            else:
                user_msg = user_msg + (
                    f"Try increasing 'max_tries_per_batch' (currently: {max_tries_per_batch}) "
                    f"or increasing 'batch_size' (currently: {batch_size}). Note that "
                    'increasing these values will also increase the sampling time.'
                )

            raise ValueError(user_msg)

        return sampled_rows

    def _sample_with_progress_bar(self, num_rows, max_tries_per_batch=100, batch_size=None,
                                  output_file_path=None, show_progress_bar=True):
        if num_rows is None:
            raise ValueError('You must specify the number of rows to sample (e.g. num_rows=100).')

        sampled = pd.DataFrame()
        if num_rows == 0:
            return sampled

        output_file_path = validate_file_path(output_file_path)
        batch_size = min(batch_size, num_rows) if batch_size else num_rows

        try:
            with tqdm.tqdm(total=num_rows, disable=not show_progress_bar) as progress_bar:
                progress_bar.set_description('Sampling rows')
                sampled = self._sample_in_batches(
                    num_rows=num_rows,
                    batch_size=batch_size,
                    max_tries_per_batch=max_tries_per_batch,
                    progress_bar=progress_bar,
                    output_file_path=output_file_path
                )

        except (Exception, KeyboardInterrupt) as error:
            handle_sampling_error(output_file_path == TMP_FILE_NAME, output_file_path, error)

        else:
            if output_file_path == TMP_FILE_NAME and os.path.exists(output_file_path):
                os.remove(output_file_path)

        return sampled

    def sample(self, num_rows, max_tries_per_batch=100, batch_size=None, output_file_path=None):
        """Sample rows from this table.

        Args:
            num_rows (int):
                Number of rows to sample. This parameter is required.
            max_tries_per_batch (int):
                Number of times to retry sampling until the batch size is met. Defaults to 100.
            batch_size (int or None):
                The batch size to sample. Defaults to ``num_rows``, if None.
            output_file_path (str or None):
                The file to periodically write sampled rows to. If None, does not
                write rows anywhere.

        Returns:
            pandas.DataFrame:
                Sampled data.
        """
        has_constraints = bool(self._data_processor._constraints)
        has_batches = batch_size is not None and batch_size != num_rows
        show_progress_bar = has_constraints or has_batches

        return self._sample_with_progress_bar(
            num_rows,
            max_tries_per_batch,
            batch_size,
            output_file_path,
            show_progress_bar=show_progress_bar
        )

    def _validate_conditions(self, conditions):
        """Validate the user-passed conditions."""
        for column in conditions.columns:
            if column not in self._data_processor.get_sdtypes():
                raise ValueError(f"Unexpected column name '{column}'. "
                                 f'Use a column name that was present in the original data.')

    def _sample_with_conditions(self, conditions, max_tries_per_batch, batch_size,
                                progress_bar=None, output_file_path=None):
        """Sample rows with conditions.

        Args:
            conditions (pandas.DataFrame):
                A DataFrame representing the conditions to be sampled.
            max_tries_per_batch (int):
                Number of times to retry sampling until the batch size is met. Defaults to 100.
            batch_size (int):
                The batch size to use for each sampling call.
            progress_bar (tqdm.tqdm or None):
                The progress bar to update.
            output_file_path (str or None):
                The file to periodically write sampled rows to. Defaults to
                a temporary file, if None.

        Returns:
            pandas.DataFrame:
                Sampled data.

        Raises:
            ConstraintsNotMetError:
                If the conditions are not valid for the given constraints.
            ValueError:
                If any of the following happens:
                    * any of the conditions' columns are not valid.
                    * no rows could be generated.
        """
        condition_columns = list(conditions.columns)
        conditions.index.name = COND_IDX
        conditions = conditions.reset_index()
        grouped_conditions = conditions.groupby(condition_columns)

        # sample
        all_sampled_rows = []

        for group, dataframe in grouped_conditions:
            if not isinstance(group, tuple):
                group = [group]

            condition = dict(zip(condition_columns, group))
            condition_df = dataframe.iloc[0].to_frame().T
            try:
                transformed_condition = self._data_processor.transform(
                    condition_df,
                    is_condition=True
                )
            except ConstraintsNotMetError as error:
                raise ConstraintsNotMetError(
                    'Provided conditions are not valid for the given constraints.'
                ) from error

            transformed_conditions = pd.concat(
                [transformed_condition] * len(dataframe),
                ignore_index=True
            )
            transformed_columns = list(transformed_conditions.columns)
            if not transformed_conditions.empty:
                transformed_conditions.index = dataframe.index
                transformed_conditions[COND_IDX] = dataframe[COND_IDX]

            if len(transformed_columns) == 0:
                sampled_rows = self._conditionally_sample_rows(
                    dataframe=dataframe,
                    condition=condition,
                    transformed_condition=None,
                    max_tries_per_batch=max_tries_per_batch,
                    batch_size=batch_size,
                    progress_bar=progress_bar,
                    output_file_path=output_file_path,
                )
                all_sampled_rows.append(sampled_rows)
            else:
                transformed_groups = transformed_conditions.groupby(transformed_columns)
                for transformed_group, transformed_dataframe in transformed_groups:
                    if not isinstance(transformed_group, tuple):
                        transformed_group = [transformed_group]

                    transformed_condition = dict(zip(transformed_columns, transformed_group))
                    sampled_rows = self._conditionally_sample_rows(
                        dataframe=transformed_dataframe,
                        condition=condition,
                        transformed_condition=transformed_condition,
                        max_tries_per_batch=max_tries_per_batch,
                        batch_size=batch_size,
                        progress_bar=progress_bar,
                        output_file_path=output_file_path,
                    )
                    all_sampled_rows.append(sampled_rows)

        all_sampled_rows = pd.concat(all_sampled_rows)
        if len(all_sampled_rows) == 0:
            return all_sampled_rows

        all_sampled_rows = all_sampled_rows.set_index(COND_IDX)
        all_sampled_rows.index.name = conditions.index.name
        all_sampled_rows = all_sampled_rows.sort_index()

        return all_sampled_rows

    def _sample_from_conditions(self, conditions, max_tries_per_batch,
                                batch_size, output_file_path):
        """Sample rows from this table with the given conditions."""
        output_file_path = validate_file_path(output_file_path)

        num_rows = functools.reduce(
            lambda num_rows, condition: condition.get_num_rows() + num_rows, conditions, 0)

        conditions = self._make_condition_dfs(conditions)
        for condition_dataframe in conditions:
            self._validate_conditions(condition_dataframe)

        sampled = pd.DataFrame()
        try:
            with tqdm.tqdm(total=num_rows) as progress_bar:
                progress_bar.set_description('Sampling conditions')
                for condition_dataframe in conditions:
                    sampled_for_condition = self._sample_with_conditions(
                        condition_dataframe,
                        max_tries_per_batch,
                        batch_size,
                        progress_bar,
                        output_file_path,
                    )
                    sampled = pd.concat([sampled, sampled_for_condition], ignore_index=True)

            is_reject_sampling = bool(
                hasattr(self, '_model') and not isinstance(self._model, GaussianMultivariate))
            check_num_rows(
                num_rows=len(sampled),
                expected_num_rows=num_rows,
                is_reject_sampling=is_reject_sampling,
                max_tries_per_batch=max_tries_per_batch
            )

        except (Exception, KeyboardInterrupt) as error:
            handle_sampling_error(output_file_path == TMP_FILE_NAME, output_file_path, error)

        else:
            if output_file_path == TMP_FILE_NAME and os.path.exists(output_file_path):
                os.remove(output_file_path)

        return sampled

    def sample_from_conditions(self, conditions, max_tries_per_batch=100,
                               batch_size=None, output_file_path=None):
        """Sample rows from this table with the given conditions.

        Args:
            conditions (list[sdv.sampling.Condition]):
                A list of sdv.sampling.Condition objects, which specify the column
                values in a condition, along with the number of rows for that
                condition.
            max_tries_per_batch (int):
                Number of times to retry sampling until the batch size is met. Defaults to 100.
            batch_size (int):
                The batch size to use per sampling call.
            output_file_path (str or None):
                The file to periodically write sampled rows to. Defaults to
                a temporary file, if None.

        Returns:
            pandas.DataFrame:
                Sampled data.

        Raises:
            ConstraintsNotMetError:
                If the conditions are not valid for the given constraints.
            ValueError:
                If any of the following happens:
                    * any of the conditions' columns are not valid.
                    * no rows could be generated.
        """
        return self._sample_from_conditions(
            conditions, max_tries_per_batch, batch_size, output_file_path)

    def _sample_remaining_columns(self, known_columns, max_tries_per_batch,
                                  batch_size, output_file_path):
        """Sample the remaining columns of a given DataFrame."""
        output_file_path = validate_file_path(output_file_path)

        known_columns = known_columns.copy()
        self._validate_conditions(known_columns)
        sampled = pd.DataFrame()
        try:
            with tqdm.tqdm(total=len(known_columns)) as progress_bar:
                progress_bar.set_description('Sampling remaining columns')
                sampled = self._sample_with_conditions(
                    known_columns, max_tries_per_batch, batch_size, progress_bar, output_file_path)

            is_reject_sampling = (hasattr(self, '_model') and not isinstance(
                self._model, copulas.multivariate.GaussianMultivariate))

            check_num_rows(
                num_rows=len(sampled),
                expected_num_rows=len(known_columns),
                is_reject_sampling=is_reject_sampling,
                max_tries_per_batch=max_tries_per_batch
            )

        except (Exception, KeyboardInterrupt) as error:
            handle_sampling_error(output_file_path == TMP_FILE_NAME, output_file_path, error)

        else:
            if output_file_path == TMP_FILE_NAME and os.path.exists(output_file_path):
                os.remove(output_file_path)

        return sampled

    def sample_remaining_columns(self, known_columns, max_tries_per_batch=100,
                                 batch_size=None, output_file_path=None):
        """Sample remaining rows from already known columns.

        Args:
            known_columns (pandas.DataFrame):
                A pandas.DataFrame with the columns that are already known. The output
                is a DataFrame such that each row in the output is sampled
                conditionally on the corresponding row in the input.
            max_tries_per_batch (int):
                Number of times to retry sampling until the batch size is met. Defaults to 100.
            batch_size (int):
                The batch size to use per sampling call.
            output_file_path (str or None):
                The file to periodically write sampled rows to. Defaults to
                a temporary file, if None.

        Returns:
            pandas.DataFrame:
                Sampled data.

        Raises:
            ConstraintsNotMetError:
                If the conditions are not valid for the given constraints.
            ValueError:
                If any of the following happens:
                    * any of the conditions' columns are not valid.
                    * no rows could be generated.
        """
        return self._sample_remaining_columns(
            known_columns,
            max_tries_per_batch,
            batch_size,
            output_file_path
        )



"""Metadata Exceptions."""


class InvalidMetadataError(Exception):
    """Error to raise when Metadata is not valid."""


class MetadataNotFittedError(InvalidMetadataError):
    """Error to raise when Metadata is used before fitting."""



"""Synthesizers for Single Table data."""

from sdv.single_table.copulagan import CopulaGANSynthesizer
from sdv.single_table.copulas import GaussianCopulaSynthesizer
from sdv.single_table.ctgan import CTGANSynthesizer, TVAESynthesizer

__all__ = (
    'GaussianCopulaSynthesizer',
    'CTGANSynthesizer',
    'TVAESynthesizer',
    'CopulaGANSynthesizer',
)



"""Base Multi Table Synthesizer class."""
import contextlib
import datetime
import warnings
from collections import defaultdict
from copy import deepcopy

import cloudpickle
import numpy as np
import pandas as pd
import pkg_resources

from sdv.errors import SynthesizerInputError
from sdv.single_table.copulas import GaussianCopulaSynthesizer
from sdv.single_table.errors import InvalidDataError


class BaseMultiTableSynthesizer:
    """Base class for multi table synthesizers.

    The ``BaseMultiTableSynthesizer`` class defines the common API that all the
    multi table synthesizers need to implement, as well as common functionality.

    Args:
        metadata (sdv.metadata.multi_table.MultiTableMetadata):
            Multi table metadata representing the data tables that this synthesizer will be used
            for.
    """

    _synthesizer = GaussianCopulaSynthesizer
    _numpy_seed = 73251

    @contextlib.contextmanager
    def _set_temp_numpy_seed(self):
        initial_state = np.random.get_state()
        if isinstance(self._numpy_seed, int):
            np.random.seed(self._numpy_seed)
            np.random.default_rng(self._numpy_seed)
        else:
            np.random.set_state(self._numpy_seed)
            np.random.default_rng(self._numpy_seed[1])
        try:
            yield
        finally:
            self._numpy_seed = np.random.get_state()
            np.random.set_state(initial_state)

    def _initialize_models(self):
        for table_name, table_metadata in self.metadata.tables.items():
            synthesizer_parameters = self._table_parameters.get(table_name, {})
            self._table_synthesizers[table_name] = self._synthesizer(
                metadata=table_metadata,
                **synthesizer_parameters
            )

    def __init__(self, metadata):
        self.metadata = metadata
        self.metadata.validate()
        self._table_synthesizers = {}
        self._table_parameters = defaultdict(dict)
        self._initialize_models()
        self._fitted = False
        self._creation_date = datetime.datetime.today().strftime('%Y-%m-%d')
        self._fitted_date = None
        self._fitted_sdv_version = None

    def get_table_parameters(self, table_name):
        """Return the parameters that will be used to instantiate the table's synthesizer.

        Args:
            table_name (str):
                Table name for which the parameters should be retrieved.

        Returns:
            parameters (dict):
                A dictionary representing the parameters that will be used to instantiate the
                table's synthesizer.
        """
        return self._table_parameters.get(table_name, {})

    def get_parameters(self, table_name):
        """Return the parameters used to instantiate the table's synthesizer.

        Args:
            table_name (str):
                Table name for which the parameters should be retrieved.

        Returns:
            parameters (dict):
                A dictionary representing the parameters used to instantiate the table's
                synthesizer.
        """
        return self._table_synthesizers.get(table_name).get_parameters()

    def set_table_parameters(self, table_name, table_parameters):
        """Update the table's synthesizer instantiation parameters.

        Args:
            table_name (str):
                Table name for which the parameters should be retrieved.
            table_parameters (dict):
                A dictionary with the parameters as keys and the values to be used to instantiate
                the table's synthesizer.
        """
        self._table_synthesizers[table_name] = self._synthesizer(
            metadata=self.metadata.tables[table_name],
            **table_parameters
        )
        self._table_parameters[table_name].update(deepcopy(table_parameters))

    def get_metadata(self):
        """Return the ``MultiTableMetadata`` for this synthesizer."""
        return self.metadata

    def _get_all_foreign_keys(self, table_name):
        foreign_keys = []
        for relation in self.metadata.relationships:
            if table_name == relation['child_table_name']:
                foreign_keys.append(deepcopy(relation['child_foreign_key']))

        return foreign_keys

    def _validate_foreign_keys(self, data):
        error_msg = None
        errors = []
        for relation in self.metadata.relationships:
            child_table = data.get(relation['child_table_name'])
            parent_table = data.get(relation['parent_table_name'])

            if isinstance(child_table, pd.DataFrame) and isinstance(parent_table, pd.DataFrame):
                child_column = child_table[relation['child_foreign_key']]
                parent_column = parent_table[relation['parent_primary_key']]
                missing_values = child_column[~child_column.isin(parent_column)].unique()

                if any(missing_values):
                    message = ', '.join(missing_values[:5].astype(str))
                    if len(missing_values) > 5:
                        message = f'({message}, + more)'
                    else:
                        message = f'({message})'

                    errors.append(
                        f"Error: foreign key column '{relation['child_foreign_key']}' contains "
                        f'unknown references: {message}. All the values in this column must '
                        'reference a primary key.'
                    )

            if errors:
                error_msg = 'Relationships:\n'
                error_msg += '\n'.join(errors)

        return error_msg

    def validate(self, data):
        """Validate data.

        Args:
            data (dict):
                A dictionary with key as table name and ``pandas.DataFrame`` as value to validate.

        Raises:
            ValueError:
                Raised when data is not of type pd.DataFrame.
            InvalidDataError:
                Raised if:
                    * foreign key does not belong to a primay key
                    * data columns don't match metadata
                    * keys have missing values
                    * primary or alternate keys are not unique
                    * context columns vary for a sequence key
                    * values of a column don't satisfy their sdtype
        """
        errors = []
        missing_tables = set(self.metadata.tables) - set(data)
        if missing_tables:
            errors.append(f'The provided data is missing the tables {missing_tables}.')

        for table_name, table_data in data.items():
            try:
                self._table_synthesizers[table_name].validate(table_data)

            except InvalidDataError as error:
                error_msg = f"Table: '{table_name}'"
                for _error in error.errors:
                    error_msg += f'\nError: {_error}'

                errors.append(error_msg)

            except ValueError as error:
                errors.append(str(error))

            except KeyError:
                continue

        foreign_key_errors = self._validate_foreign_keys(data)
        if foreign_key_errors:
            errors.append(foreign_key_errors)

        if errors:
            raise InvalidDataError(errors)

    def _validate_table_name(self, table_name):
        if table_name not in self._table_synthesizers:
            raise InvalidDataError([f"Table '{table_name}' is not present in the metadata."])

    def _assign_table_transformers(self, synthesizer, table_name, table_data):
        """Update the ``synthesizer`` to ignore the foreign keys while preprocessing the data."""
        synthesizer.auto_assign_transformers(table_data)
        foreign_key_columns = self._get_all_foreign_keys(table_name)
        column_name_to_transformers = {
            column_name: None
            for column_name in foreign_key_columns
        }
        synthesizer.update_transformers(column_name_to_transformers)

    def auto_assign_transformers(self, data):
        """Automatically assign the required transformers for the given data and constraints.

        This method will automatically set a configuration to the ``rdt.HyperTransformer``
        with the required transformers for the current data.

        Args:
            data (dict):
                Mapping of table name to pandas.DataFrame.

        Raises:
            InvalidDataError:
                If a table of the data is not present in the metadata.
        """
        for table_name, table_data in data.items():
            self._validate_table_name(table_name)
            synthesizer = self._table_synthesizers[table_name]
            self._assign_table_transformers(synthesizer, table_name, table_data)

    def get_transformers(self, table_name):
        """Get a dictionary mapping of ``column_name`` and ``rdt.transformers``.

        A dictionary representing the column names and the transformers that will be used
        to transform the data.

        Args:
            table_name (string):
                The name of the table of which to get the transformers.

        Returns:
            dict:
                A dictionary mapping with column names and transformers.

        Raises:
            ValueError:
                If ``table_name`` is not present in the metadata.
        """
        self._validate_table_name(table_name)
        return self._table_synthesizers[table_name].get_transformers()

    def update_transformers(self, table_name, column_name_to_transformer):
        """Update any of the transformers assigned to each of the column names.

        Args:
            table_name (string):
                The name of the table of which to update the transformers.
            column_name_to_transformer (dict):
                Dict mapping column names to transformers to be used for that column.

        Raises:
            ValueError:
                If ``table_name`` is not present in the metadata.
        """
        self._validate_table_name(table_name)
        self._table_synthesizers[table_name].update_transformers(column_name_to_transformer)

    def _fit(self, processed_data):
        """Fit the model to the tables.

        Args:
            processed_data (dict):
                Dictionary mapping each table name to a preprocessed ``pandas.DataFrame``.
        """
        raise NotImplementedError()

    def preprocess(self, data):
        """Transform the raw data to numerical space.

        Args:
            data (dict):
                Dictionary mapping each table name to a ``pandas.DataFrame``.

        Returns:
            dict:
                A dictionary with the preprocessed data.
        """
        self.validate(data)
        if self._fitted:
            warnings.warn(
                'This model has already been fitted. To use the new preprocessed data, '
                "please refit the model using 'fit' or 'fit_processed_data'."
            )

        processed_data = {}
        for table_name, table_data in data.items():
            synthesizer = self._table_synthesizers[table_name]
            self._assign_table_transformers(synthesizer, table_name, table_data)
            processed_data[table_name] = synthesizer._preprocess(table_data)

        return processed_data

    def fit_processed_data(self, processed_data):
        """Fit this model to the transformed data.

        Args:
            processed_data (dict):
                Dictionary mapping each table name to a preprocessed ``pandas.DataFrame``.
        """
        self._fit(processed_data.copy())
        self._fitted = True
        self._fitted_date = datetime.datetime.today().strftime('%Y-%m-%d')
        self._fitted_sdv_version = pkg_resources.get_distribution('sdv').version

    def fit(self, data):
        """Fit this model to the original data.

        Args:
            data (dict):
                Dictionary mapping each table name to a ``pandas.DataFrame`` in the raw format
                (before any transformations).
        """
        self._fitted = False
        processed_data = self.preprocess(data)
        self.fit_processed_data(processed_data)

    def reset_sampling(self):
        """Reset the sampling to the state that was left right after fitting."""
        self._numpy_seed = 73251
        for synthesizer in self._table_synthesizers.values():
            synthesizer.reset_sampling()

    def _sample(self, scale):
        raise NotImplementedError()

    def sample(self, scale=1.0):
        """Generate synthetic data for the entire dataset.

        Args:
            scale (float):
                A float representing how much to scale the data by. If scale is set to ``1.0``,
                this does not scale the sizes of the tables. If ``scale`` is greater than ``1.0``
                create more rows than the original data by a factor of ``scale``.
                If ``scale`` is lower than ``1.0`` create fewer rows by the factor of ``scale``
                than the original tables. Defaults to ``1.0``.
        """
        if not type(scale) in (float, int) or not scale > 0:
            raise SynthesizerInputError(
                f"Invalid parameter for 'scale' ({scale}). Please provide a number that is >0.0.")

        with self._set_temp_numpy_seed():
            sampled_data = self._sample(scale=scale)

        return sampled_data

    def get_learned_distributions(self, table_name):
        """Get the marginal distributions used by the ``GaussianCopula`` for a table.

        Return a dictionary mapping the column names with the distribution name and the learned
        parameters for those.

        Args:
            table_name (str):
                Table name for which the parameters should be retrieved.

        Returns:
            dict:
                Dictionary containing the distributions used or detected for each column and the
                learned parameters for those.
        """
        synthesizer = self._table_synthesizers[table_name]
        return synthesizer.get_learned_distributions()

    def _validate_constraints(self, constraints):
        for constraint_dict in constraints:
            if 'table_name' not in constraint_dict.keys():
                raise SynthesizerInputError(
                    "A constraint is missing required parameter 'table_name'. "
                    'Please add this parameter to your constraint definition.'
                )

            if constraint_dict['constraint_class'] == 'Unique':
                raise SynthesizerInputError(
                    "The constraint class 'Unique' is not currently supported for multi-table"
                    ' synthesizers. Please remove the constraint for this synthesizer.'
                )

        if self._fitted:
            warnings.warn(
                "For these constraints to take effect, please refit the synthesizer using 'fit'."
            )

    def add_constraints(self, constraints):
        """Add constraints to the synthesizer.

        Args:
            constraints (list):
                List of constraints described as dictionaries in the following format:
                    * ``constraint_class``: Name of the constraint to apply.
                    * ``table_name``: Name of the table where to apply the constraint.
                    * ``constraint_parameters``: A dictionary with the constraint parameters.

        Raises:
            SynthesizerInputError:
                Raises when the ``Unique`` constraint is passed.
        """
        self._validate_constraints(constraints)
        for constraint in constraints:
            constraint = deepcopy(constraint)
            synthesizer = self._table_synthesizers[constraint.pop('table_name')]
            synthesizer._data_processor.add_constraints([constraint])

    def get_constraints(self):
        """Get constraints of the synthesizer.

        Returns:
            list:
                List of dictionaries describing the constraints of the synthesizer.
        """
        constraints = []
        for table_name, synthesizer in self._table_synthesizers.items():
            for constraint in synthesizer.get_constraints():
                constraint['table_name'] = table_name
                constraints.append(constraint)

        return constraints

    def load_custom_constraint_classes(self, table_name, filepath, class_names):
        """Load a custom constraint class for the specified table's synthesizer.

        Args:
            table_name (str):
                Table to add constraint to.
            filepath (str):
                String representing the absolute or relative path to the python file where
                the custom constraints are declared.
            class_names (list):
                A list of custom constraint classes to be imported.
        """
        self._table_synthesizers[table_name].load_custom_constraint_classes(filepath, class_names)

    def add_custom_constraint_class(self, table_name, class_object, class_name):
        """Add a custom constraint class for the synthesizer to use.

        Args:
            table_name (str):
                Table to add constraint to.
            class_object (sdv.constraints.Constraint):
                A custom constraint class object.
            class_name (str):
                The name to assign this custom constraint class. This will be the name to use
                when writing a constraint dictionary for ``add_constraints``.
        """
        self._table_synthesizers[table_name].add_custom_constraint_class(class_object, class_name)

    def get_info(self):
        """Get dictionary with information regarding the synthesizer.

        Return:
            dict:
                * ``class_name``: synthesizer class name
                * ``creation_date``: date of creation
                * ``is_fit``: whether or not the synthesizer has been fit
                * ``last_fit_date``: date for the last time it was fit
                * ``fitted_sdv_version``: version of sdv it was on when fitted
        """
        return {
            'class_name': self.__class__.__name__,
            'creation_date': self._creation_date,
            'is_fit': self._fitted,
            'last_fit_date': self._fitted_date,
            'fitted_sdv_version': self._fitted_sdv_version
        }

    def save(self, filepath):
        """Save this instance to the given path using cloudpickle.

        Args:
            filepath (str):
                Path where the instance will be serialized.
        """
        with open(filepath, 'wb') as output:
            cloudpickle.dump(self, output)

    @classmethod
    def load(cls, path):
        """Load the multi-table synthesizer from a given path.

        Args:
            path (str):
                Path from which to load the instance.
        """
        with open(path, 'rb') as f:
            synthesizer = cloudpickle.load(f)
            return synthesizer



from unittest.mock import Mock, patch

from sdv.metadata.single_table import SingleTableMetadata
from sdv.single_table.ctgan import CTGANSynthesizer, TVAESynthesizer


class TestCTGANSynthesizer:

    def test___init__(self):
        """Test creating an instance of ``CTGANSynthesizer``."""
        # Setup
        metadata = SingleTableMetadata()
        enforce_min_max_values = True
        enforce_rounding = True

        # Run
        instance = CTGANSynthesizer(
            metadata,
            enforce_min_max_values=enforce_min_max_values,
            enforce_rounding=enforce_rounding,
        )

        # Assert
        assert instance.enforce_min_max_values is True
        assert instance.enforce_rounding is True
        assert instance.embedding_dim == 128
        assert instance.generator_dim == (256, 256)
        assert instance.discriminator_dim == (256, 256)
        assert instance.generator_lr == 2e-4
        assert instance.generator_decay == 1e-6
        assert instance.discriminator_lr == 2e-4
        assert instance.discriminator_decay == 1e-6
        assert instance.batch_size == 500
        assert instance.discriminator_steps == 1
        assert instance.log_frequency is True
        assert instance.verbose is False
        assert instance.epochs == 300
        assert instance.pac == 10
        assert instance.cuda is True

    def test___init__custom(self):
        """Test creating an instance of ``CTGANSynthesizer`` with custom parameters."""
        # Setup
        metadata = SingleTableMetadata()
        enforce_min_max_values = False
        enforce_rounding = False
        embedding_dim = 64
        generator_dim = (128, 128)
        discriminator_dim = (128, 128)
        generator_lr = 1e-4
        generator_decay = 2e-6
        discriminator_lr = 3e-4
        discriminator_decay = 1e-6
        batch_size = 250
        discriminator_steps = 2
        log_frequency = False
        verbose = True
        epochs = 150
        pac = 5
        cuda = False

        # Run
        instance = CTGANSynthesizer(
            metadata,
            enforce_min_max_values=enforce_min_max_values,
            enforce_rounding=enforce_rounding,
            embedding_dim=embedding_dim,
            generator_dim=generator_dim,
            discriminator_dim=discriminator_dim,
            generator_lr=generator_lr,
            generator_decay=generator_decay,
            discriminator_lr=discriminator_lr,
            discriminator_decay=discriminator_decay,
            batch_size=batch_size,
            discriminator_steps=discriminator_steps,
            log_frequency=log_frequency,
            verbose=verbose,
            epochs=epochs,
            pac=pac,
            cuda=cuda,
        )

        # Assert
        assert instance.enforce_min_max_values is False
        assert instance.enforce_rounding is False
        assert instance.embedding_dim == embedding_dim
        assert instance.generator_dim == generator_dim
        assert instance.discriminator_dim == discriminator_dim
        assert instance.generator_lr == generator_lr
        assert instance.generator_decay == generator_decay
        assert instance.discriminator_lr == discriminator_lr
        assert instance.discriminator_decay == discriminator_decay
        assert instance.batch_size == batch_size
        assert instance.discriminator_steps == discriminator_steps
        assert instance.log_frequency == log_frequency
        assert instance.verbose is True
        assert instance.epochs == epochs
        assert instance.pac == pac
        assert instance.cuda is False

    def test_get_parameters(self):
        """Test that inherited method ``get_parameters`` returns the specific init parameters."""
        # Setup
        metadata = SingleTableMetadata()
        instance = CTGANSynthesizer(metadata)

        # Run
        result = instance.get_parameters()

        # Assert
        assert result == {
            'enforce_min_max_values': True,
            'enforce_rounding': True,
            'embedding_dim': 128,
            'generator_dim': (256, 256),
            'discriminator_dim': (256, 256),
            'generator_lr': 2e-4,
            'generator_decay': 1e-6,
            'discriminator_lr': 2e-4,
            'discriminator_decay': 1e-6,
            'batch_size': 500,
            'discriminator_steps': 1,
            'log_frequency': True,
            'verbose': False,
            'epochs': 300,
            'pac': 10,
            'cuda': True,
        }

    @patch('sdv.single_table.ctgan.CTGAN')
    @patch('sdv.single_table.ctgan.detect_discrete_columns')
    def test__fit(self, mock_detect_discrete_columns, mock_ctgan):
        """Test the ``_fit`` from ``CTGANSynthesizer``.

        Test that when we call ``_fit`` a new instance of ``CTGAN`` is created as a model
        with the previously set parameters. Then this is being fitted with the ``discrete_columns``
        that have been detected by the utility function.
        """
        # Setup
        metadata = SingleTableMetadata()
        instance = CTGANSynthesizer(metadata)
        processed_data = Mock()

        # Run
        instance._fit(processed_data)

        # Assert
        mock_detect_discrete_columns.assert_called_once_with(metadata, processed_data)
        mock_ctgan.assert_called_once_with(
            batch_size=500,
            cuda=True,
            discriminator_decay=1e-6,
            discriminator_dim=(256, 256),
            discriminator_lr=2e-4,
            discriminator_steps=1,
            embedding_dim=128,
            epochs=300,
            generator_decay=1e-6,
            generator_dim=(256, 256),
            generator_lr=2e-4,
            log_frequency=True,
            pac=10,
            verbose=False,
        )

        assert instance._model == mock_ctgan.return_value
        instance._model.fit.assert_called_once_with(
            processed_data,
            discrete_columns=mock_detect_discrete_columns.return_value
        )


class TestTVAESynthesizer:

    def test___init__(self):
        """Test creating an instance of ``TVAESynthesizer``."""
        # Setup
        metadata = SingleTableMetadata()
        enforce_min_max_values = True
        enforce_rounding = True

        # Run
        instance = TVAESynthesizer(
            metadata,
            enforce_min_max_values=enforce_min_max_values,
            enforce_rounding=enforce_rounding,
        )

        # Assert
        assert instance.enforce_min_max_values is True
        assert instance.enforce_rounding is True
        assert instance.embedding_dim == 128
        assert instance.compress_dims == (128, 128)
        assert instance.decompress_dims == (128, 128)
        assert instance.l2scale == 1e-5
        assert instance.batch_size == 500
        assert instance.epochs == 300
        assert instance.loss_factor == 2
        assert instance.cuda is True

    def test___init__custom(self):
        """Test creating an instance of ``TVAESynthesizer`` with custom parameters."""
        # Setup
        metadata = SingleTableMetadata()
        enforce_min_max_values = False
        enforce_rounding = False
        embedding_dim = 64
        compress_dims = (64, 64)
        decompress_dims = (64, 64)
        l2scale = 2e-5
        batch_size = 250
        epochs = 150
        loss_factor = 4
        cuda = False

        # Run
        instance = TVAESynthesizer(
            metadata,
            enforce_min_max_values=enforce_min_max_values,
            enforce_rounding=enforce_rounding,
            embedding_dim=embedding_dim,
            compress_dims=compress_dims,
            decompress_dims=decompress_dims,
            l2scale=l2scale,
            batch_size=batch_size,
            epochs=epochs,
            loss_factor=loss_factor,
            cuda=cuda,
        )

        # Assert
        assert instance.enforce_min_max_values is False
        assert instance.enforce_rounding is False
        assert instance.embedding_dim == 64
        assert instance.compress_dims == (64, 64)
        assert instance.decompress_dims == (64, 64)
        assert instance.l2scale == 2e-5
        assert instance.batch_size == 250
        assert instance.epochs == 150
        assert instance.loss_factor == 4
        assert instance.cuda is False

    def test_get_parameters(self):
        """Test that inherited method ``get_parameters`` returns the specific init parameters."""
        # Setup
        metadata = SingleTableMetadata()
        instance = TVAESynthesizer(metadata)

        # Run
        result = instance.get_parameters()

        # Assert
        assert result == {
            'enforce_min_max_values': True,
            'enforce_rounding': True,
            'embedding_dim': 128,
            'compress_dims': (128, 128),
            'decompress_dims': (128, 128),
            'l2scale': 1e-5,
            'batch_size': 500,
            'epochs': 300,
            'loss_factor': 2,
            'cuda': True,
        }

    @patch('sdv.single_table.ctgan.TVAE')
    @patch('sdv.single_table.ctgan.detect_discrete_columns')
    def test__fit(self, mock_detect_discrete_columns, mock_tvae):
        """Test the ``_fit`` from ``TVAESynthesizer``.

        Test that when we call ``_fit`` a new instance of ``TVAE`` is created as a model
        with the previously set parameters. Then this is being fitted with the ``discrete_columns``
        that have been detected by the utility function.
        """
        # Setup
        metadata = SingleTableMetadata()
        instance = TVAESynthesizer(metadata)
        processed_data = Mock()

        # Run
        instance._fit(processed_data)

        # Assert
        mock_detect_discrete_columns.assert_called_once_with(metadata, processed_data)
        mock_tvae.assert_called_once_with(
            batch_size=500,
            compress_dims=(128, 128),
            cuda=True,
            decompress_dims=(128, 128),
            embedding_dim=128,
            epochs=300,
            l2scale=1e-5,
            loss_factor=2,
        )

        assert instance._model == mock_tvae.return_value
        instance._model.fit.assert_called_once_with(
            processed_data,
            discrete_columns=mock_detect_discrete_columns.return_value
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.md', encoding='utf-8') as readme_file:
    readme = readme_file.read()

with open('HISTORY.md', encoding='utf-8') as history_file:
    history = history_file.read()


install_requires = [
    'Faker>=10,<15',
    'graphviz>=0.13.2,<1',
    "numpy>=1.20.0,<2;python_version<'3.10'",
    "numpy>=1.23.3,<2;python_version>='3.10'",
    "pandas>=1.1.3,<2;python_version<'3.10'",
    "pandas>=1.5.0,<2;python_version>='3.10'",
    'tqdm>=4.15,<5',
    'copulas>=0.8.0,<0.9',
    'ctgan>=0.7.1,<0.8',
    'deepecho>=0.4.0,<0.5',
    'rdt>=1.3.0,<2',
    'sdmetrics>=0.9.1,<0.10',
    'cloudpickle>=2.1.0,<3.0',
    'boto3>=1.15.0,<2',
    'botocore>=1.18,<2'
]

pomegranate_requires = [
    "pomegranate>=0.14.3,<0.15",
]

setup_requires = [
    'pytest-runner>=2.11.1',
]

tests_require = [
    'pytest>=3.4.2',
    'pytest-cov>=2.6.0',
    'pytest-rerunfailures>10',
    'jupyter>=1.0.0,<2',
    'rundoc>=0.4.3,<0.5',
]

development_requires = [
    # general
    'bumpversion>=0.5.3,<0.6',
    'pip>=9.0.1',
    'watchdog>=0.8.3,<0.9',

    # docs
    'docutils>=0.12,<0.18',
    'm2r2>=0.2.5,<0.3',
    'nbsphinx>=0.5.0,<0.7',
    'sphinx_toolbox>=2.5,<3',
    'Sphinx>=3,<3.3',
    'pydata-sphinx-theme<0.5',
    'markupsafe<2.1.0',

    # Jinja2>=3 makes the sphinx theme fail
    'Jinja2>=2,<3',

    # style check
    'flake8>=3.7.7,<4',
    'flake8-absolute-import>=1.0,<2',
    'flake8-builtins>=1.5.3,<1.6',
    'flake8-comprehensions>=3.6.1,<3.7',
    'flake8-debugger>=4.0.0,<4.1',
    'flake8-docstrings>=1.5.0,<2',
    'flake8-eradicate>=1.1.0,<1.2',
    'flake8-fixme>=1.1.1,<1.2',
    'flake8-mock>=0.3,<0.4',
    'flake8-multiline-containers>=0.0.18,<0.1',
    'flake8-mutable>=1.2.0,<1.3',
    'flake8-expression-complexity>=0.0.9,<0.1',
    'flake8-print>=4.0.0,<4.1',
    'flake8-pytest-style>=1.5.0,<2',
    'flake8-quotes>=3.3.0,<4',
    'flake8-sfs>=0.0.3,<0.1',
    'flake8-variables-names>=0.0.4,<0.1',
    'dlint>=0.11.0,<0.12',
    'isort>=4.3.4,<5',
    'pandas-vet>=0.2.3,<0.3',
    'pep8-naming>=0.12.1,<0.13',
    'pydocstyle>=6.1.1,<6.2',

    # fix style issues
    'autoflake>=1.1,<2',
    'autopep8>=1.4.3,<1.6',

    # distribute on PyPI
    'twine>=1.10.0,<4',
    'wheel>=0.30.0',

    # Advanced testing
    'coverage>=4.5.1,<6',
    'tox>=2.9.1,<4',
    'invoke'
]

setup(
    author='DataCebo, Inc.',
    author_email='info@sdv.dev',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: Free for non-commercial use',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
    description='Generate synthetic data for single table, multi table and sequential data',
    extras_require={
        'test': tests_require,
        'dev': development_requires + tests_require,
        'pomegranate': pomegranate_requires,
    },
    include_package_data=True,
    install_requires=install_requires,
    keywords='sdv synthetic-data synhtetic-data-generation timeseries single-table multi-table',
    license='BSL-1.1',
    long_description=readme + '\n\n' + history,
    long_description_content_type='text/markdown',
    name='sdv',
    packages=find_packages(include=['sdv', 'sdv.*']),
    python_requires='>=3.7,<3.11',
    setup_requires=setup_requires,
    test_suite='tests',
    tests_require=tests_require,
    url='https://github.com/sdv-dev/SDV',
    version='1.0.0',
    zip_safe=False,
)



import datetime

import numpy as np
import pandas as pd
import pkg_resources
import pytest
from faker import Faker

from sdv.datasets.demo import download_demo
from sdv.datasets.local import load_csvs
from sdv.evaluation.multi_table import evaluate_quality, get_column_pair_plot, get_column_plot
from sdv.metadata.multi_table import MultiTableMetadata
from sdv.multi_table import HMASynthesizer
from tests.integration.single_table.custom_constraints import MyConstraint


def test_hma():
    """End to end integration tests with ``HMASynthesizer``.

    The test consist on loading the demo data, convert the old metadata to the new format
    and then fit a ``HMASynthesizer``. After fitting two samples are being generated, one with
    a 0.5 scale and one with 1.5 scale.
    """
    # Setup
    data, metadata = download_demo('multi_table', 'got_families')
    hmasynthesizer = HMASynthesizer(metadata)

    # Run
    hmasynthesizer.fit(data)
    normal_sample = hmasynthesizer.sample(0.5)
    increased_sample = hmasynthesizer.sample(1.5)

    # Assert
    assert list(normal_sample) == ['characters', 'character_families', 'families']
    assert list(increased_sample) == ['characters', 'character_families', 'families']
    for table_name, table in normal_sample.items():
        assert all(table.columns == data[table_name].columns)

    for normal_table, increased_table in zip(normal_sample.values(), increased_sample.values()):
        assert increased_table.size > normal_table.size


def test_hma_reset_sampling():
    """End to end integration test that uses ``reset_sampling``.

    This test uses ``reset_sampling`` to ensure that the model will generate the same data
    as the first sample after this method has been called.
    """
    # Setup
    faker = Faker()
    data, metadata = download_demo('multi_table', 'got_families')
    metadata.add_column(
        'characters',
        'ssn',
        sdtype='ssn',
    )
    data['characters']['ssn'] = [faker.lexify() for _ in range(len(data['characters']))]
    for table in metadata.tables.values():
        table.alternate_keys = []

    hmasynthesizer = HMASynthesizer(metadata)

    # Run
    hmasynthesizer.fit(data)
    first_sample = hmasynthesizer.sample()
    second_sample = hmasynthesizer.sample()
    hmasynthesizer.reset_sampling()
    reset_first_sample = hmasynthesizer.sample()
    reset_second_sample = hmasynthesizer.sample()

    # Assert
    for table, reset_table in zip(first_sample.values(), reset_first_sample.values()):
        pd.testing.assert_frame_equal(table, reset_table)

    for table, reset_table in zip(second_sample.values(), reset_second_sample.values()):
        pd.testing.assert_frame_equal(table, reset_table)

    for sample_1, sample_2 in zip(first_sample.values(), second_sample.values()):
        with pytest.raises(AssertionError):
            pd.testing.assert_frame_equal(sample_1, sample_2)


def test_get_info():
    """Test the correct dictionary is returned.

    Check the return dictionary is valid both before and after fitting the synthesizer.
    """
    # Setup
    data = {'tab': pd.DataFrame({'col': [1, 2, 3]})}
    today = datetime.datetime.today().strftime('%Y-%m-%d')
    metadata = MultiTableMetadata()
    metadata.add_table('tab')
    metadata.add_column('tab', 'col', sdtype='numerical')
    synthesizer = HMASynthesizer(metadata)

    # Run
    info = synthesizer.get_info()

    # Assert
    assert info == {
        'class_name': 'HMASynthesizer',
        'creation_date': today,
        'is_fit': False,
        'last_fit_date': None,
        'fitted_sdv_version': None
    }

    # Run
    synthesizer.fit(data)
    info = synthesizer.get_info()

    # Assert
    version = pkg_resources.get_distribution('sdv').version
    assert info == {
        'class_name': 'HMASynthesizer',
        'creation_date': today,
        'is_fit': True,
        'last_fit_date': today,
        'fitted_sdv_version': version
    }


def test_hma_set_parameters():
    """Test the ``set_table_parameters``.

    Validate that the ``set_table_parameters`` sets new parameters to the synthesizers.
    """
    # Setup
    data, metadata = download_demo('multi_table', 'got_families')
    hmasynthesizer = HMASynthesizer(metadata)

    # Run
    hmasynthesizer.set_table_parameters('characters', {'default_distribution': 'gamma'})
    hmasynthesizer.set_table_parameters('families', {'default_distribution': 'uniform'})
    hmasynthesizer.set_table_parameters('character_families', {'default_distribution': 'norm'})

    # Assert
    assert hmasynthesizer.get_table_parameters('characters') == {'default_distribution': 'gamma'}
    assert hmasynthesizer.get_table_parameters('families') == {'default_distribution': 'uniform'}
    assert hmasynthesizer.get_table_parameters('character_families') == {
        'default_distribution': 'norm'
    }

    assert hmasynthesizer._table_synthesizers['characters'].default_distribution == 'gamma'
    assert hmasynthesizer._table_synthesizers['families'].default_distribution == 'uniform'
    assert hmasynthesizer._table_synthesizers['character_families'].default_distribution == 'norm'


def get_custom_constraint_data_and_metadata():
    parent_data = pd.DataFrame({
        'primary_key': [1000, 1001, 1002],
        'numerical_col': [2, 3, 4],
        'categorical_col': ['a', 'b', 'a'],
    })
    child_data = pd.DataFrame({
        'user_id': [1000, 1001, 1000],
        'id': [1, 2, 3],
        'random': ['a', 'b', 'c']
    })

    metadata = MultiTableMetadata()
    metadata.detect_table_from_dataframe('parent', parent_data)
    metadata.update_column('parent', 'primary_key', sdtype='id')
    metadata.detect_table_from_dataframe('child', child_data)
    metadata.update_column('child', 'user_id', sdtype='id')
    metadata.update_column('child', 'id', sdtype='id')
    metadata.set_primary_key('parent', 'primary_key')
    metadata.set_primary_key('child', 'id')
    metadata.add_relationship(
        parent_primary_key='primary_key',
        parent_table_name='parent',
        child_foreign_key='user_id',
        child_table_name='child'
    )

    return parent_data, child_data, metadata


def test_hma_custom_constraint():
    """Test an example of using a custom constraint."""
    parent_data, child_data, metadata = get_custom_constraint_data_and_metadata()
    synthesizer = HMASynthesizer(metadata)
    constraint = {
        'table_name': 'parent',
        'constraint_class': 'MyConstraint',
        'constraint_parameters': {
            'column_names': ['numerical_col']
        }
    }
    synthesizer.add_custom_constraint_class('parent', MyConstraint, 'MyConstraint')

    # Run
    synthesizer.add_constraints(constraints=[constraint])
    processed_data = synthesizer.preprocess({'parent': parent_data, 'child': child_data})

    # Assert Processed Data
    np.testing.assert_equal(
        processed_data['parent']['numerical_col'].array,
        (parent_data['numerical_col'] ** 2.0).array
    )

    # Run - Fit the model
    synthesizer.fit_processed_data(processed_data)

    # Run - sample
    sampled = synthesizer.sample(10)
    assert all(sampled['parent']['numerical_col'] > 1)


def test_hma_custom_constraint_loaded_from_file():
    """Test an example of using a custom constraint loaded from a file."""
    parent_data, child_data, metadata = get_custom_constraint_data_and_metadata()
    synthesizer = HMASynthesizer(metadata)
    constraint = {
        'table_name': 'parent',
        'constraint_class': 'MyConstraint',
        'constraint_parameters': {
            'column_names': ['numerical_col']
        }
    }
    synthesizer.load_custom_constraint_classes(
        'parent',
        'tests/integration/single_table/custom_constraints.py',
        ['MyConstraint']
    )

    # Run
    synthesizer.add_constraints(constraints=[constraint])
    processed_data = synthesizer.preprocess({'parent': parent_data, 'child': child_data})

    # Assert Processed Data
    np.testing.assert_equal(
        processed_data['parent']['numerical_col'].array,
        (parent_data['numerical_col'] ** 2.0).array
    )

    # Run - Fit the model
    synthesizer.fit_processed_data(processed_data)

    # Run - sample
    sampled = synthesizer.sample(10)
    assert all(sampled['parent']['numerical_col'] > 1)


def test_hma_with_inequality_constraint():
    """Test that ensures that when new columns are created by the constraint this still works."""
    # Setup
    parent_table = pd.DataFrame(data={
        'id': [1, 2, 3, 4, 5],
        'column': [1.2, 2.1, 2.2, 2.1, 1.4]
    })

    child_table = pd.DataFrame(data={
        'id': [1, 2, 3, 4, 5],
        'parent_id': [1, 1, 3, 2, 1],
        'low_column': [1, 3, 3, 1, 2],
        'high_column': [2, 4, 5, 2, 4]
    })

    data = {
        'parent_table': parent_table,
        'child_table': child_table
    }

    metadata = MultiTableMetadata()
    metadata.detect_table_from_dataframe(table_name='parent_table', data=parent_table)
    metadata.update_column('parent_table', 'id', sdtype='id')
    metadata.detect_table_from_dataframe(table_name='child_table', data=child_table)
    metadata.update_column('child_table', 'id', sdtype='id')
    metadata.update_column('child_table', 'parent_id', sdtype='id')

    metadata.set_primary_key(table_name='parent_table', column_name='id')
    metadata.set_primary_key(table_name='child_table', column_name='id')

    metadata.add_relationship(
        parent_table_name='parent_table',
        child_table_name='child_table',
        parent_primary_key='id',
        child_foreign_key='parent_id'
    )

    constraint = {
        'constraint_class': 'Inequality',
        'table_name': 'child_table',
        'constraint_parameters': {
            'low_column_name': 'low_column',
            'high_column_name': 'high_column'
        }
    }

    synthesizer = HMASynthesizer(metadata)

    # Run
    synthesizer.add_constraints(constraints=[constraint])
    synthesizer.fit(data)
    sampled = synthesizer.sample(10)

    # Assert
    assert all(sampled['child_table']['low_column'] < sampled['child_table']['high_column'])


def test_fit_processed_multiple_calls():
    """Test that ``fit_processed_data`` does not modify input data."""
    # Setup
    parent_data, child_data, metadata = get_custom_constraint_data_and_metadata()
    synthesizer = HMASynthesizer(metadata)

    # Run
    preprocessed = synthesizer.preprocess({'parent': parent_data, 'child': child_data})
    parent_copy = preprocessed['parent'].copy()
    child_copy = preprocessed['child'].copy()

    synthesizer.fit_processed_data(preprocessed)

    # Assert
    assert preprocessed.keys() == {'parent', 'child'}
    pd.testing.assert_frame_equal(parent_copy, preprocessed['parent'], check_like=True)
    pd.testing.assert_frame_equal(child_copy, preprocessed['child'], check_like=True)

    # Re-run to ensure it does not error
    synthesizer.fit_processed_data(preprocessed)


def test_save_and_load(tmp_path):
    """Test saving and loading a multi-table synthesizer."""
    # Setup
    _, _, metadata = get_custom_constraint_data_and_metadata()
    synthesizer = HMASynthesizer(metadata)
    model_path = tmp_path / 'synthesizer.pkl'

    # Run
    synthesizer.save(model_path)

    # Assert
    assert model_path.exists()
    assert model_path.is_file()
    loaded_synthesizer = HMASynthesizer.load(model_path)

    assert isinstance(synthesizer, HMASynthesizer)
    assert loaded_synthesizer.get_info() == synthesizer.get_info()
    assert loaded_synthesizer.metadata.to_dict() == metadata.to_dict()


def test_hma_primary_key_and_foreign_key_only():
    """Test that ``HMASynthesizer`` can handle tables with primary and foreign keys only."""
    # Setup
    users = pd.DataFrame({
        'user_id': [1, 2, 3],
        'user_name': ['John', 'Doe', 'Johanna']
    })
    sessions = pd.DataFrame({
        'session_id': ['a', 'b', 'c'],
        'clicks': [10, 20, 30]
    })
    games = pd.DataFrame({
        'game_id': ['a1', 'b2', 'c3'],
        'session_id': ['a', 'b', 'c'],
        'user_id': [1, 2, 3]
    })

    data = {
        'users': users,
        'sessions': sessions,
        'games': games
    }

    metadata = MultiTableMetadata()
    for table_name, table in data.items():
        metadata.detect_table_from_dataframe(table_name, table)

    metadata.update_column('users', 'user_id', sdtype='id')
    metadata.update_column('sessions', 'session_id', sdtype='id')
    metadata.update_column('games', 'game_id', sdtype='id')
    metadata.update_column('games', 'session_id', sdtype='id')
    metadata.update_column('games', 'user_id', sdtype='id')
    metadata.set_primary_key('users', 'user_id')
    metadata.set_primary_key('sessions', 'session_id')
    metadata.set_primary_key('games', 'game_id')
    metadata.add_relationship('users', 'games', 'user_id', 'user_id')
    metadata.add_relationship('sessions', 'games', 'session_id', 'session_id')

    hmasynthesizer = HMASynthesizer(metadata)

    # Fit
    hmasynthesizer.fit(data)

    # Sample
    sample = hmasynthesizer.sample()

    # Assert
    assert all(sample['games']['user_id'].isin(sample['users']['user_id']))
    assert all(sample['games']['session_id'].isin(sample['sessions']['session_id']))


def test_synthesize_multiple_tables_using_hma(tmp_path):
    """End to end test for multiple tables using ``HMASynthesizer``.

    The following functionalities are being tested:
        * Create, fit and sample from model
        * Anonymization
        * Evaluating synthetic data
        * Saving, loading and sampling from the loaded model
        * Using a custom configuration for the ``HMASynthesizer``
    """

    # Loading the demo data
    real_data, metadata = download_demo(
        modality='multi_table',
        dataset_name='fake_hotels'
    )

    # Creating a Synthesizer
    synthesizer = HMASynthesizer(metadata)
    synthesizer.fit(real_data)

    # Generating Synthetic Data
    synthetic_data = synthesizer.sample(scale=2)

    # Assert new data is bigger than real_data
    for table_name in metadata.tables:
        assert len(synthetic_data[table_name]) > len(real_data[table_name])

    # Assert Anonymization
    sensitive_columns = ['guest_email', 'billing_address', 'credit_card_number']
    for column in sensitive_columns:
        assert synthetic_data['guests'][column].isin(real_data['guests'][column]).sum() == 0

    # Evaluate Real vs Synthetic Data
    quality_report = evaluate_quality(
        real_data,
        synthetic_data,
        metadata,
        verbose=False
    )

    column_plot = get_column_plot(
        real_data=real_data,
        synthetic_data=synthetic_data,
        column_name='has_rewards',
        table_name='guests',
        metadata=metadata
    )

    column_pair_plot = get_column_pair_plot(
        real_data=real_data,
        synthetic_data=synthetic_data,
        column_names=['room_rate', 'room_type'],
        table_name='guests',
        metadata=metadata
    )

    # Assert
    assert quality_report.get_score() > 0
    assert column_plot
    assert column_pair_plot

    # Save and Load
    model_path = tmp_path / 'synthesizer.pkl'

    synthesizer.save(model_path)

    # Assert
    assert model_path.exists()
    assert model_path.is_file()
    loaded_synthesizer = HMASynthesizer.load(model_path)

    assert isinstance(synthesizer, HMASynthesizer)
    assert loaded_synthesizer.get_info() == synthesizer.get_info()
    assert loaded_synthesizer.metadata.to_dict() == metadata.to_dict()
    loaded_synthesizer.sample()

    # HMA Customization
    custom_synthesizer = HMASynthesizer(
        metadata
    )

    custom_synthesizer.set_table_parameters(
        table_name='hotels',
        table_parameters={
            'default_distribution': 'truncnorm'
        }
    )

    custom_synthesizer.fit(real_data)
    learned_distributions = custom_synthesizer.get_learned_distributions(table_name='hotels')

    # Assert
    assert list(learned_distributions['rating']['learned_parameters']) == [
        'a',
        'b',
        'loc',
        'scale'
    ]
    assert learned_distributions['rating']['distribution'] == 'truncnorm'


def test_use_own_data_using_hma(tmp_path):
    """End to end test for loading and preparing data.

    Tests loading data from CSVs, auto-detecting metadata from the dataframes, updating the
    metadata with correct sdtypes, and finally fitting the synthesizer.
    """
    # Setup
    data_folder = tmp_path / 'datasets'
    download_demo(
        modality='multi_table',
        dataset_name='fake_hotels',
        output_folder_name=data_folder
    )

    # Run - load CSVs
    datasets = load_csvs(data_folder)

    # Assert - loaded CSVs correctly
    assert datasets.keys() == {'guests', 'hotels'}

    # Metadata
    metadata = MultiTableMetadata()

    metadata.detect_table_from_dataframe(
        table_name='guests',
        data=datasets['guests']
    )
    metadata.detect_table_from_dataframe(
        table_name='hotels',
        data=datasets['hotels']
    )

    # Assert - detected metadata correctly
    for table in metadata.tables:
        assert metadata.tables[table].columns.keys() == set(datasets[table].columns)

    # Update metadata
    metadata.update_column(
        table_name='guests',
        column_name='checkin_date',
        sdtype='datetime',
        datetime_format='%d %b %Y'
    )
    metadata.update_column(
        table_name='guests',
        column_name='checkout_date',
        sdtype='datetime',
        datetime_format='%d %b %Y'
    )
    metadata.update_column(
        table_name='hotels',
        column_name='hotel_id',
        sdtype='id',
        regex_format='HID_[0-9]{3,4}'
    )
    metadata.update_column(
        table_name='guests',
        column_name='hotel_id',
        sdtype='id',
        regex_format='HID_[0-9]{3,4}'
    )
    metadata.update_column(
        table_name='guests',
        column_name='guest_email',
        sdtype='email',
        pii=True
    )
    metadata.update_column(
        table_name='guests',
        column_name='billing_address',
        sdtype='address',
        pii=True
    )
    metadata.update_column(
        table_name='guests',
        column_name='credit_card_number',
        sdtype='credit_card_number',
        pii=True
    )
    metadata.set_primary_key(
        table_name='hotels',
        column_name='hotel_id'
    )
    metadata.set_primary_key(
        table_name='guests',
        column_name='guest_email'
    )
    metadata.add_alternate_keys(
        table_name='guests',
        column_names=['credit_card_number']
    )
    metadata.add_relationship(
        parent_table_name='hotels',
        child_table_name='guests',
        parent_primary_key='hotel_id',
        child_foreign_key='hotel_id'
    )

    # Assert - check updated metadata
    metadata.validate()
    hotels_metadata = metadata.tables['hotels']
    assert hotels_metadata.primary_key == 'hotel_id'
    assert hotels_metadata.columns['hotel_id']['sdtype'] == 'id'
    assert hotels_metadata.columns['hotel_id']['regex_format'] == 'HID_[0-9]{3,4}'

    guests_metadata = metadata.tables['guests']
    assert guests_metadata.primary_key == 'guest_email'
    assert guests_metadata.alternate_keys == ['credit_card_number']
    assert guests_metadata.columns['checkin_date']['sdtype'] == 'datetime'
    assert guests_metadata.columns['checkin_date']['datetime_format'] == '%d %b %Y'
    assert guests_metadata.columns['checkout_date']['sdtype'] == 'datetime'
    assert guests_metadata.columns['checkout_date']['datetime_format'] == '%d %b %Y'
    assert guests_metadata.columns['hotel_id']['sdtype'] == 'id'
    assert guests_metadata.columns['hotel_id']['regex_format'] == 'HID_[0-9]{3,4}'
    assert guests_metadata.columns['guest_email']['sdtype'] == 'email'
    assert guests_metadata.columns['guest_email']['pii'] is True
    assert guests_metadata.columns['billing_address']['sdtype'] == 'address'
    assert guests_metadata.columns['billing_address']['pii'] is True
    assert guests_metadata.columns['credit_card_number']['sdtype'] == 'credit_card_number'
    assert guests_metadata.columns['credit_card_number']['pii'] is True

    # Save and load metadata
    metadata_path = tmp_path / 'metadata.json'
    metadata.save_to_json(metadata_path)
    loaded_metadata = MultiTableMetadata.load_from_json(metadata_path)

    # Assert loaded metadata matches saved
    assert metadata.to_dict() == loaded_metadata.to_dict()

    # Fit synthesizer
    synthesizer = HMASynthesizer(metadata)
    synthesizer.validate(datasets)
    synthesizer.fit(datasets)
    synthetic_data = synthesizer.sample(scale=1)
    synthesizer.validate(synthetic_data)

    for table in metadata.tables:
        assert set(synthetic_data[table].columns) == set(datasets[table].columns)



import re
from unittest.mock import Mock

import numpy as np
import pandas as pd
import pytest

from sdv.multi_table.hma import HMASynthesizer
from sdv.single_table.copulas import GaussianCopulaSynthesizer
from tests.utils import get_multi_table_data, get_multi_table_metadata


class TestHMASynthesizer:

    def test___init__(self):
        """Test the default initialization of the ``HMASynthesizer``."""
        # Run
        metadata = get_multi_table_metadata()
        metadata.validate = Mock()
        instance = HMASynthesizer(metadata)

        # Assert
        assert instance.metadata == metadata
        assert isinstance(instance._table_synthesizers['nesreca'], GaussianCopulaSynthesizer)
        assert isinstance(instance._table_synthesizers['oseba'], GaussianCopulaSynthesizer)
        assert isinstance(instance._table_synthesizers['upravna_enota'], GaussianCopulaSynthesizer)
        assert instance._table_parameters == {
            'nesreca': {'default_distribution': 'beta'},
            'oseba': {'default_distribution': 'beta'},
            'upravna_enota': {'default_distribution': 'beta'},
        }
        instance.metadata.validate.assert_called_once_with()

    def test__get_extension(self):
        """Test the ``_get_extension`` method.

        Test that the resulting dataframe contains extended columns using the names
        and parameters from a trained ``copulas.univariate`` model.
        """
        # Setup
        metadata = get_multi_table_metadata()
        child_table = pd.DataFrame({
            'id_nesreca': [0, 1, 2, 3],
            'upravna_enota': [0, 1, 2, 3]
        })
        instance = HMASynthesizer(metadata)

        # Run
        result = instance._get_extension('nesreca', child_table, 'upravna_enota')

        # Assert
        expected = pd.DataFrame({
            '__nesreca__upravna_enota__univariates__id_nesreca__a': [1., 1., 1., 1.],
            '__nesreca__upravna_enota__univariates__id_nesreca__b': [1., 1., 1., 1.],
            '__nesreca__upravna_enota__univariates__id_nesreca__loc': [0., 1., 2., 3.],
            '__nesreca__upravna_enota__univariates__id_nesreca__scale': [np.nan] * 4,
            '__nesreca__upravna_enota__num_rows': [1., 1., 1., 1.]
        })

        pd.testing.assert_frame_equal(result, expected)

    def test__get_extension_foreign_key_only(self):
        """Test the ``_get_extension`` method.

        Test when foreign key only is passed, just the ``num_rows`` is being captured.
        """
        # Setup
        instance = Mock()
        instance._get_all_foreign_keys.return_value = ['id_upravna_enota']
        instance._table_synthesizers = {'nesreca': Mock()}
        child_table = pd.DataFrame({
            'id_upravna_enota': [0, 1, 2, 3]
        })

        # Run
        result = HMASynthesizer._get_extension(
            instance,
            'nesreca',
            child_table,
            'id_upravna_enota'
        )

        # Assert
        expected = pd.DataFrame({
            '__nesreca__id_upravna_enota__num_rows': [1, 1, 1, 1]
        })

        pd.testing.assert_frame_equal(result, expected)

    def test__get_foreign_keys(self):
        """Test that this method returns the foreign keys for a given table name and child name."""
        # Setup
        metadata = get_multi_table_metadata()
        instance = HMASynthesizer(metadata)

        # Run
        result = instance._get_foreign_keys('nesreca', 'oseba')

        # Assert
        assert result == ['id_nesreca']

    def test__get_all_foreign_keys(self):
        """Test that this method returns all the foreign keys for a given table name."""
        # Setup
        metadata = get_multi_table_metadata()
        instance = HMASynthesizer(metadata)

        # Run
        result = instance._get_all_foreign_keys('nesreca')

        # Assert
        assert result == ['upravna_enota']

    def test__extend_table(self):
        """Test that ``extend_table`` extends the current table with extra columns.

        This also updates ``self._modeled_tables`` and ``self._max_child_rows``.
        """
        # Setup
        metadata = get_multi_table_metadata()
        instance = HMASynthesizer(metadata)
        metadata.add_column('nesreca', 'value', sdtype='numerical')
        metadata.add_column('oseba', 'oseba_value', sdtype='numerical')
        metadata.add_column('upravna_enota', 'name', sdtype='categorical')

        data = get_multi_table_data()
        data['nesreca']['value'] = [0, 1, 2, 3]
        data['oseba']['oseba_value'] = [0, 1, 2, 3]

        # Run
        result = instance._extend_table(data['nesreca'], data, 'nesreca')

        # Assert
        expected_result = pd.DataFrame({
            'id_nesreca': [0, 1, 2, 3],
            'upravna_enota': [0, 1, 2, 3],
            'nesreca_val': [0, 1, 2, 3],
            'value': [0, 1, 2, 3],
            '__oseba__id_nesreca__covariance__0__0': [0.] * 4,
            '__oseba__id_nesreca__univariates__oseba_val__a': [1.] * 4,
            '__oseba__id_nesreca__univariates__oseba_val__b': [1.] * 4,
            '__oseba__id_nesreca__univariates__oseba_val__loc': [0., 1., 2., 3.],
            '__oseba__id_nesreca__univariates__oseba_val__scale': [np.nan] * 4,
            '__oseba__id_nesreca__univariates__oseba_value__a': [1.] * 4,
            '__oseba__id_nesreca__univariates__oseba_value__b': [1.] * 4,
            '__oseba__id_nesreca__univariates__oseba_value__loc': [0., 1., 2., 3.],
            '__oseba__id_nesreca__univariates__oseba_value__scale': [np.nan] * 4,
            '__oseba__id_nesreca__num_rows': [1.] * 4,
        })

        pd.testing.assert_frame_equal(expected_result, result)
        assert instance._modeled_tables == ['oseba']
        assert instance._max_child_rows['__oseba__id_nesreca__num_rows'] == 1

    def test__pop_foreign_keys(self):
        """Test that this method removes the foreign keys from the ``table_data``."""
        # Setup
        instance = Mock()
        instance._get_all_foreign_keys.return_value = ['a', 'b']
        table_data = pd.DataFrame({
            'a': [1, 2, 3],
            'b': [2, 3, 4],
            'c': ['John', 'Doe', 'Johanna']
        })

        # Run
        result = HMASynthesizer._pop_foreign_keys(instance, table_data, 'table_name')

        # Assert
        pd.testing.assert_frame_equal(pd.DataFrame({'c': ['John', 'Doe', 'Johanna']}), table_data)
        np.testing.assert_array_equal(result['a'], [1, 2, 3])
        np.testing.assert_array_equal(result['b'], [2, 3, 4])

    def test__clear_nans(self):
        """Test that this method clears all the nans and substitutes them with expected values."""
        # Setup
        data = pd.DataFrame({
            'numerical': [0, 1, 2, 3, np.nan, np.nan],
            'categorical': ['John', np.nan, 'Johanna', 'John', np.nan, 'Doe'],
        })

        # Run
        HMASynthesizer._clear_nans(data)

        # Assert
        expected_data = pd.DataFrame({
            'numerical': [0, 1, 2, 3, 1.5, 1.5],
            'categorical': ['John', 'John', 'Johanna', 'John', 'John', 'Doe']
        })
        pd.testing.assert_frame_equal(expected_data, data)

    def test__model_table(self):
        """Test that ``_model_table`` performs the modeling.

        Modeling consists of getting the table for the given table name,
        learning the size of this table, removing the foreign keys and clearing
        any null values by using the ``_clear_nans`` method. Then, fitting the table model by
        calling ``fit_processed_data``,  adding back the foreign keys, updating the ``tables`` and
        marking the table name as modeled within the ``instance._modeled_tables``.
        """
        # Setup
        nesreca_model = Mock()
        instance = Mock()
        instance._synthesizer = GaussianCopulaSynthesizer

        instance._modeled_tables = []
        instance._table_sizes = {}
        instance._table_synthesizers = {'nesreca': nesreca_model}
        instance._pop_foreign_keys.return_value = {'fk': [1, 2, 3]}
        data = pd.DataFrame({
            'id_nesreca': [0, 1, 2],
            'upravna_enota': [0, 1, 2]
        })
        extended_data = pd.DataFrame({
            'id_nesreca': [0, 1, 2],
            'upravna_enota': [0, 1, 2],
            'extended': ['a', 'b', 'c']
        })

        instance._extend_table.return_value = extended_data

        tables = {'nesreca': data}

        # Run
        result = HMASynthesizer._model_table(instance, 'nesreca', tables)

        # Assert
        expected_result = pd.DataFrame({
            'id_nesreca': [0, 1, 2],
            'upravna_enota': [0, 1, 2],
            'extended': ['a', 'b', 'c'],
            'fk': [1, 2, 3]
        })
        pd.testing.assert_frame_equal(expected_result, result)

        instance._extend_table.assert_called_once_with(data, tables, 'nesreca')
        instance._pop_foreign_keys.assert_called_once_with(extended_data, 'nesreca')
        instance._clear_nans(extended_data)
        nesreca_model.fit_processed_data.assert_called_once_with(extended_data)

        assert instance._modeled_tables == ['nesreca']
        assert instance._table_sizes == {'nesreca': 3}

    def test__fit(self):
        """Test that ``_fit`` calls ``_model_table`` only if the table has no parents."""
        # Setup
        metadata = get_multi_table_metadata()
        instance = HMASynthesizer(metadata)
        instance._model_table = Mock()
        data = get_multi_table_data()
        data['nesreca']['value'] = [0, 1, 2, 3]
        data['oseba']['oseba_value'] = [0, 1, 2, 3]

        # Run
        instance._fit(data)

        # Assert
        instance._model_table.assert_called_once_with('upravna_enota', data)

    def test__finalize(self):
        """Test that the finalize method applies the final touches to the generated data.

        The process consists of applying the propper data types to each table, and finding
        foreign keys if those are not present in the current sampled data.
        """
        # Setup
        instance = Mock()
        metadata = Mock()
        metadata._get_parent_map.return_value = {
            'sessions': ['users'],
            'transactions': ['sessions']
        }
        instance.metadata = metadata

        instance._get_foreign_keys.side_effect = [['user_id'], ['session_id']]
        instance._find_parent_ids.return_value = pd.Series(['a', 'a', 'b'], name='session_id')

        sampled_data = {
            'users': pd.DataFrame({
                'user_id': [0, 1, 2],
                'name': ['John', 'Doe', 'Johanna'],
                'additional_column': [0.1, 0.2, 0.3],
                'another_additional_column': [0.1, 0.2, 0.5]
            }, dtype=np.int64),
            'sessions': pd.DataFrame({
                'user_id': [1, 2, 1],
                'session_id': ['a', 'b', 'c'],
                'os': ['linux', 'mac', 'win'],
                'country': ['us', 'us', 'es']
            }, dtype=np.int64),
            'transactions': pd.DataFrame({
                'transaction_id': [1, 2, 3],
            }, dtype=np.int64),
        }

        users_synth = Mock()
        users_synth._data_processor._dtypes = {'user_id': np.int64, 'name': str}
        sessions_synth = Mock()
        sessions_synth._data_processor._dtypes = {
            'user_id': np.int64,
            'session_id': str,
            'os': str,
            'country': str
        }
        transactions_synth = Mock()
        transactions_synth._data_processor._dtypes = {
            'transaction_id': np.int64,
            'session_id': str
        }

        instance._table_synthesizers = {
            'users': users_synth,
            'sessions': sessions_synth,
            'transactions': transactions_synth
        }

        # Run
        result = HMASynthesizer._finalize(instance, sampled_data)

        # Assert
        expected_result = {
            'users': pd.DataFrame({
                'user_id': [0, 1, 2],
                'name': ['John', 'Doe', 'Johanna'],
            }, dtype=np.int64),
            'sessions': pd.DataFrame({
                'user_id': [1, 2, 1],
                'session_id': ['a', 'b', 'c'],
                'os': ['linux', 'mac', 'win'],
                'country': ['us', 'us', 'es'],
            }, dtype=np.int64),
            'transactions': pd.DataFrame({
                'transaction_id': [1, 2, 3],
                'session_id': ['a', 'a', 'b']
            }, dtype=np.int64),
        }
        for result_frame, expected_frame in zip(result.values(), expected_result.values()):
            pd.testing.assert_frame_equal(result_frame, expected_frame)

    def test__extract_parameters(self):
        """Test that parameters are being returned without the prefix."""
        # Setup
        parent_row = pd.Series({
            '__sessions__user_id__num_rows': 10,
            '__sessions__user_id__a': 1.0,
            '__sessions__user_id__b': 0.2,
            '__sessions__user_id__loc': 0.5,
            '__sessions__user_id__scale': 0.25
        })
        instance = Mock()
        instance._max_child_rows = {'__sessions__user_id__num_rows': 10}

        # Run
        result = HMASynthesizer._extract_parameters(instance, parent_row, 'sessions', 'user_id')

        # Assert
        expected_result = {
            'a': 1.0,
            'b': 0.2,
            'loc': 0.5,
            'num_rows': 10.0,
            'scale': 0.25
        }

        assert result == expected_result

    def test__process_samples(self):
        """Test the ``_process_samples``.

        Test that the method retrieves the ``data_processor`` from the fitted ``table_synthesizer``
        and performs a ``reverse_transform`` and returns the data in the real space.
        """
        # Setup
        sampled_rows = pd.DataFrame({
            'name': [0.1, 0.25, 0.35],
            'a': [1.0, 0.25, 0.5],
            'b': [0.2, 0.6, 0.9],
            'loc': [0.5, 0.1, 0.2],
            'num_rows': [1, 2, 3],
            'scale': [0.25, 0.35, 0.15]
        })
        instance = Mock()
        users_synthesizer = Mock()
        users_synthesizer._data_processor.reverse_transform.return_value = pd.DataFrame({
            'user_id': [0, 1, 2],
            'name': ['John', 'Doe', 'Johanna']
        })
        instance._table_synthesizers = {'users': users_synthesizer}

        # Run
        result = HMASynthesizer._process_samples(instance, 'users', sampled_rows)

        # Assert
        expected_result = pd.DataFrame({
            'user_id': [0, 1, 2],
            'name': ['John', 'Doe', 'Johanna'],
            'a': [1.0, 0.25, 0.5],
            'b': [0.2, 0.6, 0.9],
            'loc': [0.5, 0.1, 0.2],
            'num_rows': [1, 2, 3],
            'scale': [0.25, 0.35, 0.15]
        })
        result = result.reindex(sorted(result.columns), axis=1)
        expected_result = expected_result.reindex(sorted(expected_result.columns), axis=1)
        pd.testing.assert_frame_equal(result, expected_result)

    def test__sample_rows(self):
        """Test sample rows.

        Test that sampling rows will return the reverse transformed data with the extension columns
        sampled by the model.
        """
        # Setup
        synthesizer = Mock()
        instance = Mock()

        # Run
        result = HMASynthesizer._sample_rows(instance, synthesizer, 'users', 10)

        # Assert
        assert result == instance._process_samples.return_value
        instance._process_samples.assert_called_once_with(
            'users',
            synthesizer._sample.return_value
        )
        synthesizer._sample.assert_called_once_with(10)

    def test__get_child_synthesizer(self):
        """Test that this method returns a synthesizer for the given child table."""
        # Setup
        instance = Mock()
        parent_row = 'row'
        table_name = 'users'
        foreign_key = 'session_id'
        table_meta = Mock()
        instance.metadata.tables = {'users': table_meta}
        instance._synthesizer_kwargs = {'a': 1}

        # Run
        synthesizer = HMASynthesizer._get_child_synthesizer(
            instance,
            parent_row,
            table_name,
            foreign_key
        )

        # Assert
        assert synthesizer == instance._synthesizer.return_value
        instance._synthesizer.assert_called_once_with(table_meta, a=1)
        synthesizer._set_parameters.assert_called_once_with(
            instance._extract_parameters.return_value
        )
        instance._extract_parameters.assert_called_once_with(parent_row, table_name, foreign_key)

    def test__sample_child_rows(self):
        """Test the sampling of child rows when sampled data is empty."""
        # Setup
        instance = Mock()
        instance._get_foreign_keys.return_value = ['user_id']
        instance._extract_parameters.return_value = {
            'a': 1.0,
            'b': 0.2,
            'loc': 0.5,
            'num_rows': 10.0,
            'scale': 0.25
        }

        metadata = Mock()
        sessions_meta = Mock()
        users_meta = Mock()
        users_meta.primary_key.return_value = 'user_id'
        metadata.tables = {
            'users': users_meta,
            'sessions': sessions_meta
        }
        instance.metadata = metadata
        instance._synthesizer_kwargs = {'a': 0.1, 'b': 0.5, 'loc': 0.25}

        instance._sample_rows.return_value = pd.DataFrame({
            'session_id': ['a', 'b', 'c'],
            'os': ['linux', 'mac', 'win'],
            'country': ['us', 'us', 'es'],
        })
        parent_row = pd.DataFrame({
            'user_id': [1, 2, 3],
            'name': ['John', 'Doe', 'Johanna']
        })
        sampled_data = {}

        # Run
        HMASynthesizer._sample_child_rows(instance, 'sessions', 'users', parent_row, sampled_data)

        # Assert
        expected_result = pd.DataFrame({
            'session_id': ['a', 'b', 'c'],
            'os': ['linux', 'mac', 'win'],
            'country': ['us', 'us', 'es'],
            'user_id': [1, 2, 3]
        })
        pd.testing.assert_frame_equal(sampled_data['sessions'], expected_result)

    def test__sample_child_rows_with_sampled_data(self):
        """Test the sampling of child rows when sampled data contains values.

        The new sampled data has to be concatenated to the current sampled data.
        """
        # Setup
        instance = Mock()
        instance._get_foreign_keys.return_value = ['user_id']
        instance._extract_parameters.return_value = {
            'a': 1.0,
            'b': 0.2,
            'loc': 0.5,
            'num_rows': 10.0,
            'scale': 0.25
        }

        metadata = Mock()
        sessions_meta = Mock()
        users_meta = Mock()
        users_meta.primary_key.return_value = 'user_id'
        metadata.tables = {
            'users': users_meta,
            'sessions': sessions_meta
        }
        instance.metadata = metadata
        instance._synthesizer_kwargs = {'a': 0.1, 'b': 0.5, 'loc': 0.25}

        instance._sample_rows.return_value = pd.DataFrame({
            'session_id': ['a', 'b', 'c'],
            'os': ['linux', 'mac', 'win'],
            'country': ['us', 'us', 'es'],
        })
        parent_row = pd.DataFrame({
            'user_id': [1, 2, 3],
            'name': ['John', 'Doe', 'Johanna']
        })
        sampled_data = {
            'sessions': pd.DataFrame({
                'user_id': [0, 1, 0],
                'session_id': ['d', 'e', 'f'],
                'os': ['linux', 'mac', 'win'],
                'country': ['us', 'us', 'es'],
            })
        }

        # Run
        HMASynthesizer._sample_child_rows(instance, 'sessions', 'users', parent_row, sampled_data)

        # Assert
        expected_result = pd.DataFrame({
            'user_id': [0, 1, 0, 1, 2, 3],
            'session_id': ['d', 'e', 'f', 'a', 'b', 'c'],
            'os': ['linux', 'mac', 'win', 'linux', 'mac', 'win'],
            'country': ['us', 'us', 'es', 'us', 'us', 'es'],
        })
        pd.testing.assert_frame_equal(sampled_data['sessions'], expected_result)

    def test__sample_children(self):
        """Test that child tables are being sampled recursively."""
        # Setup
        def update_sampled_data(child_name, table_name, row, sampled_data):
            sampled_data['sessions'] = pd.DataFrame({
                'user_id': [1],
                'session_id': ['d'],
                'os': ['linux'],
                'country': ['us'],
            })

        metadata = Mock()
        metadata._get_child_map.return_value = {'users': ['sessions']}
        instance = Mock()
        table_rows = pd.DataFrame({
            'user_id': [1, 2, 3],
            'name': ['John', 'Doe', 'Johanna']
        })
        sampled_data = {}
        instance.metadata = metadata
        instance._sample_child_rows.side_effect = update_sampled_data

        # Run
        HMASynthesizer._sample_children(instance, 'users', sampled_data, table_rows)

        # Assert
        assert instance._sample_child_rows.call_count == 3
        sample_calls = instance._sample_child_rows.call_args_list
        pd.testing.assert_series_equal(
            sample_calls[0][0][2],
            pd.Series({'user_id': 1, 'name': 'John'}, name=0)
        )
        pd.testing.assert_series_equal(
            sample_calls[1][0][2],
            pd.Series({'user_id': 2, 'name': 'Doe'}, name=1)
        )
        pd.testing.assert_series_equal(
            sample_calls[2][0][2],
            pd.Series({'user_id': 3, 'name': 'Johanna'}, name=2)
        )

    def test__sample_table(self):
        """Test sampling a table.

        The ``sample_table`` method will call sample children and return the sampled data
        dictionary.
        """
        # Setup
        def sample_children(table_name, sampled_data, table_rows):
            sampled_data['sessions'] = pd.DataFrame({
                'user_id': [1, 1, 3],
                'session_id': ['a', 'b', 'c'],
                'os': ['windows', 'linux', 'mac'],
                'country': ['us', 'us', 'es']
            })
            sampled_data['transactions'] = pd.DataFrame({
                'transaction_id': [1, 2, 3],
                'session_id': ['a', 'a', 'b']
            })

        instance = Mock()
        instance._table_sizes = {'users': 10}
        instance._table_synthesizers = {'users': Mock()}
        instance._sample_children.side_effect = sample_children
        instance._sample_rows.return_value = pd.DataFrame({
            'user_id': [1, 2, 3],
            'name': ['John', 'Doe', 'Johanna']
        })

        # Run
        result = HMASynthesizer._sample_table(instance, 'users')

        # Assert
        expected_result = {
            'users': pd.DataFrame({
                'user_id': [1, 2, 3],
                'name': ['John', 'Doe', 'Johanna'],
            }),
            'sessions': pd.DataFrame({
                'user_id': [1, 1, 3],
                'session_id': ['a', 'b', 'c'],
                'os': ['windows', 'linux', 'mac'],
                'country': ['us', 'us', 'es'],
            }),
            'transactions': pd.DataFrame({
                'transaction_id': [1, 2, 3],
                'session_id': ['a', 'a', 'b']
            })
        }
        for result_frame, expected_frame in zip(result.values(), expected_result.values()):
            pd.testing.assert_frame_equal(result_frame, expected_frame)

    def test__sample(self):
        """Test that the ``_sample_table`` is called for tables that don't have parents."""
        # Setup
        instance = Mock()
        instance.metadata._get_parent_map.return_value = {
            'sessions': ['users'],
            'transactions': ['sessions']
        }
        instance.metadata.tables = {
            'users': Mock(),
            'sessions': Mock(),
            'transactions': Mock(),
        }

        # Run
        result = HMASynthesizer._sample(instance)

        # Assert
        assert result == instance._finalize.return_value
        instance._sample_table.assert_called_once_with('users', scale=1, sampled_data={})
        instance._finalize.assert_called_once_with({})

    def test_get_learned_distributions(self):
        """Test that ``get_learned_distributions`` returns a dict.

        Test that it returns a dictionary with the name of the columns and the learned
        distribution and its parameters.
        """
        # Setup
        metadata = get_multi_table_metadata()
        instance = HMASynthesizer(metadata)
        data = {
            'nesreca': pd.DataFrame({
                'id_nesreca': np.arange(10),
                'upravna_enota': np.arange(10),
                'nesreca_val': np.arange(10),
            }),
            'oseba': pd.DataFrame({
                'upravna_enota': np.arange(10),
                'id_nesreca': np.arange(10),
                'oseba_val': np.arange(10),
            }),
            'upravna_enota': pd.DataFrame({
                'id_upravna_enota': np.arange(10),
                'upravna_val': [10, np.nan] * 5,
            }),
        }

        instance.fit(data)

        # Run
        result = instance.get_learned_distributions('upravna_enota')

        # Assert
        assert list(result) == ['upravna_val']
        assert result['upravna_val'] == {
            'distribution': 'beta',
            'learned_parameters': {
                'a': 1.0,
                'b': 1.0,
                'loc': 10.0,
                'scale': 0.0
            }
        }

    def test_get_learned_distributions_raises_an_error(self):
        """Test that ``get_learned_distributions`` raises an error."""
        # Setup
        metadata = get_multi_table_metadata()
        metadata.add_column('nesreca', 'value', sdtype='numerical')
        metadata.add_column('oseba', 'value', sdtype='numerical')
        metadata.add_column('upravna_enota', 'a_value', sdtype='numerical')
        instance = HMASynthesizer(metadata)

        # Run and Assert
        error_msg = re.escape(
            "Distributions have not been learned yet. Please fit your model first using 'fit'."
        )
        with pytest.raises(ValueError, match=error_msg):
            instance.get_learned_distributions('upravna_enota')



"""Metadata module."""

from sdv.metadata import visualization
from sdv.metadata.errors import InvalidMetadataError, MetadataNotFittedError
from sdv.metadata.multi_table import MultiTableMetadata
from sdv.metadata.single_table import SingleTableMetadata

__all__ = (
    'InvalidMetadataError',
    'MetadataNotFittedError',
    'MultiTableMetadata',
    'SingleTableMetadata',
    'visualization'
)



"""Base class for single table model presets."""

import logging
import sys

import cloudpickle
import rdt.transformers

from sdv.single_table import GaussianCopulaSynthesizer

LOGGER = logging.getLogger(__name__)

FAST_ML_PRESET = 'FAST_ML'
PRESETS = {
    FAST_ML_PRESET: 'Use this preset to minimize the time needed to create a synthetic data model.'
}


class SingleTablePreset:
    """Class for all single table synthesizer presets.

    Args:
        metadata (sdv.metadata.SingleTableMetadata):
            ``SingleTableMetadata`` instance.
        name (str):
            The preset to use.
    """

    _synthesizer = None
    _default_synthesizer = GaussianCopulaSynthesizer

    def _setup_fast_preset(self, metadata):
        self._synthesizer = GaussianCopulaSynthesizer(
            metadata=metadata,
            default_distribution='norm',
            enforce_rounding=False
        )
        self._synthesizer._data_processor._update_transformers_by_sdtypes(
            'categorical',
            rdt.transformers.FrequencyEncoder(add_noise=True)
        )

    def __init__(self, metadata, name):
        if name not in PRESETS:
            raise ValueError(f"'name' must be one of {PRESETS}.")

        self.name = name
        if name == FAST_ML_PRESET:
            self._setup_fast_preset(metadata)

    def add_constraints(self, constraints):
        """Add constraints to the synthesizer.

        Args:
            constraints (list):
                List of constraints described as dictionaries in the following format:
                    * ``constraint_class``: Name of the constraint to apply.
                    * ``constraint_parameters``: A dictionary with the constraint parameters.
        """
        self._synthesizer.add_constraints(constraints)

    def get_metadata(self):
        """Return the ``SingleTableMetadata`` for this synthesizer."""
        return self._synthesizer.get_metadata()

    def get_parameters(self):
        """Return the parameters used to instantiate the synthesizer."""
        return self._synthesizer.get_parameters()

    def fit(self, data):
        """Fit this model to the data.

        Args:
            data (pandas.DataFrame):
                Data to fit the model to.
        """
        self._synthesizer.fit(data)

    def sample(self, num_rows, max_tries_per_batch=100, batch_size=None, output_file_path=None):
        """Sample rows from this table.

        Args:
            num_rows (int):
                Number of rows to sample. This parameter is required.
            max_tries_per_batch (int):
                Number of times to try sampling discarded rows. Defaults to 100.
            batch_size (int or None):
                The batch size to sample. Defaults to `num_rows`, if None.
            output_file_path (str or None):
                The file to periodically write sampled rows to. If None, does not
                write rows anywhere.

        Returns:
            pandas.DataFrame:
                Sampled data.
        """
        sampled = self._synthesizer.sample(
            num_rows,
            max_tries_per_batch,
            batch_size,
            output_file_path,
        )

        return sampled

    def sample_from_conditions(self, conditions, max_tries_per_batch=100,
                               batch_size=None, output_file_path=None):
        """Sample rows from this table with the given conditions.

        Args:
            conditions (list[sdv.sampling.Condition]):
                A list of sdv.sampling.Condition objects, which specify the column
                values in a condition, along with the number of rows for that
                condition.
            max_tries_per_batch (int):
                Number of times to try sampling discarded rows. Defaults to 100.
            batch_size (int):
                The batch size to use per attempt at sampling. Defaults to 10 times
                the number of rows.
            output_file_path (str or None):
                The file to periodically write sampled rows to. Defaults to
                a temporary file, if None.

        Returns:
            pandas.DataFrame:
                Sampled data.
        """
        sampled = self._synthesizer.sample_from_conditions(
            conditions,
            max_tries_per_batch,
            batch_size,
            output_file_path
        )

        return sampled

    def sample_remaining_columns(self, known_columns, max_tries_per_batch=100,
                                 batch_size=None, output_file_path=None):
        """Sample rows from this table.

        Args:
            known_columns (pandas.DataFrame):
                A pandas.DataFrame with the columns that are already known. The output
                is a DataFrame such that each row in the output is sampled
                conditionally on the corresponding row in the input.
            max_tries_per_batch (int):
                Number of times to try sampling discarded rows. Defaults to 100.
            batch_size (int):
                The batch size to use per attempt at sampling. Defaults to 10 times
                the number of rows.
            output_file_path (str or None):
                The file to periodically write sampled rows to. Defaults to
                a temporary file, if None.

        Returns:
            pandas.DataFrame:
                Sampled data.
        """
        sampled = self._synthesizer.sample_remaining_columns(
            known_columns,
            max_tries_per_batch,
            batch_size,
            output_file_path
        )

        return sampled

    def save(self, path):
        """Save this model instance to the given path using cloudpickle.

        Args:
            path (str):
                Path where the SDV instance will be serialized.
        """
        with open(path, 'wb') as output:
            cloudpickle.dump(self, output)

    @classmethod
    def load(cls, path):
        """Load a SingleTableSynthesizer instance from a given path.

        Args:
            path (str):
                Path from which to load the instance.

        Returns:
            SingleTableSynthesizer:
                The loaded synthesizer.
        """
        with open(path, 'rb') as f:
            model = cloudpickle.load(f)
            return model

    @classmethod
    def list_available_presets(cls, out=sys.stdout):
        """List the available presets and their descriptions."""
        out.write(f'Available presets:\n{PRESETS}\n\n'
                  'Supply the desired preset using the `name` parameter.\n\n'
                  'Have any requests for custom presets? Contact the SDV team to learn '
                  'more an SDV Premium license.\n')

    def __repr__(self):
        """Represent single table preset instance as text.

        Returns:
            str
        """
        return f'SingleTablePreset(name={self.name})'



"""Constraint Exceptions."""


class MissingConstraintColumnError(Exception):
    """Error used when constraint is provided a table with missing columns."""

    def __init__(self, missing_columns):
        self.missing_columns = missing_columns


class AggregateConstraintsError(Exception):
    """Error used to represent a list of constraint errors."""

    def __init__(self, errors):
        self.errors = errors

    def __str__(self):
        return '\n' + '\n\n'.join(map(str, self.errors))


class InvalidFunctionError(Exception):
    """Error used when an invalid function is utilized."""


class FunctionError(Exception):
    """Error used when an a function produces an unexpected error."""


class ConstraintMetadataError(Exception):
    """Error to raise when Metadata is not valid."""
